from types import MappingProxyType

from ..basic import EnsureInt
from ..exceptions import (
    CommandParametrizationError,
    ConstraintError,
    ConstraintErrors,
    ParameterConstraintContext,
    ParameterContextErrors,
    ParametrizationErrors,
)


def test_constrainterror_repr():
    c = EnsureInt()
    ce = ConstraintError(c, 'noint', 'yeah, bullshit')
    assert repr(ce) == \
        f"ConstraintError({c!r}, 'noint', 'yeah, bullshit', None)"


def test_constrainterrors():
    c = EnsureInt()
    ce = ConstraintError(c, 'noint', 'yeah, bullshit')
    emap = dict(c1=ce)
    ces = ConstraintErrors(emap)
    assert ces.errors == emap
    assert isinstance(ces.errors, MappingProxyType)
    assert repr(ces) == f"ConstraintErrors({emap!r})"


def test_parametercontext():
    assert str(ParameterConstraintContext(('p1',))) == 'Context<p1>'
    assert str(ParameterConstraintContext(
        ('p1', 'p2'),
        'some details',
    )) == 'Context<p1, p2 (some details)>'


def test_parametercontexterrors():
    c = EnsureInt()
    ce = ConstraintError(c, 'noint', 'yeah, bullshit')
    emap = {
        ParameterConstraintContext(('c1',)): ce,
    }
    pces = ParameterContextErrors(emap)
    assert pces.items() == emap.items()
    assert repr(pces) == repr(emap)


def test_parameterizationerrors():
    c = EnsureInt()
    ce = ConstraintError(c, 'noint', 'yeah, bullshit')
    emap = {
        ParameterConstraintContext(('c1',)): ce,
    }
    pes = ParametrizationErrors(emap)
    assert str(pes) == """\
1 parameter constraint violation
c1='noint'
  yeah, bullshit"""

    # CommandParametrizationError is pretty much the same thing
    cpes = CommandParametrizationError(emap)
    assert str(cpes) == """\
1 command parameter constraint violation
c1='noint'
  yeah, bullshit"""



"""Constraints for command/function parameters"""

from __future__ import annotations

from collections.abc import Container
from itertools import chain
from typing import (
    Callable,
    Dict,
)

from .base import Constraint
from .basic import (
    NoConstraint,
)
from .dataset import DatasetParameter
from .exceptions import (
    ConstraintError,
    ParametrizationErrors,
    CommandParametrizationError,
    ParameterConstraintContext,
)


class NoValue:
    """Type to annotate the absence of a value

    For example in a list of parameter defaults. In general `None` cannot
    be used, as it may be an actual value, hence we use a local, private
    type.
    """
    pass


class EnsureCommandParameterization(Constraint):
    """Base class for `ValidatedInterface` parameter validators

    This class can be used as-is, by declaring individual constraints
    in the constructor, or it can be subclassed to consolidate all
    custom validation-related code for a command in a single place.

    Commonly this constraint is used by declaring particular value constraints
    for individual parameters as a mapping. Declaring that the ``path``
    parameter should receive something that is or can be coerced to
    a valid ``Path`` object looks like this::

      EnsureCommandParameterization({'path': EnsurePath()})

    This class differs from a standard ``Constraint`` implementation,
    because its ``__call__()`` method support additional arguments
    that are used by the internal ``Interface`` handling code to
    control how parameters are validated.

    During validation, when no validator for a particular parameter is
    declared, any input value is passed on as-is, and otherwise an input is
    passed through the validator.

    There is one exception to this rule: When a parameter value is identical to
    its default value (as declared in the command signature, and communicated
    via the ``at_default`` argument of ``__call__()``), this default
    value is also passed as-is, unless the respective parameter name is
    included in the ``validate_defaults`` constructor argument.

    An important consequence of this behavior is that validators need
    not cover a default value. For example, a parameter constraint for
    ``path=None``, where ``None`` is a special value used to indicate an
    optional and unset value, but actually only paths are acceptable input
    values. can simply use ``EnsurePath()`` and it is not necessary to do
    something like ``EnsurePath() | EnsureNone()``.

    However, `EnsureCommandParameterization` can also be specifically
    instructed to perform validation of defaults for individual parameters, as
    described above.  A common use case is the auto-discovery of datasets,
    where often `None` is the default value of a `dataset` parameter (to make
    it optional), and an `EnsureDataset` constraint is used. This constraint
    can perform the auto-discovery (with the `None` value indicating that), but
    validation of defaults must be turned on for the `dataset` parameter in
    order to do that.

    A second difference to a common ``Constraint`` implementation is the
    ability to perform an "exhaustive validation" on request (via
    ``__call__(on_error=...)``). In this case, validation is not stopped at the
    first discovered violation, but all violations are collected and
    communicated by raising a ``CommandParametrizationError`` exception, which
    can be inspected by a caller for details on number and nature of all
    discovered violations.

    Exhaustive validation and joint reporting are only supported for individual
    constraint implementations that raise `ConstraintError` exceptions. For
    legacy constraints, any raised exception of another type are not caught
    and reraised immediately.
    """
    def __init__(
        self,
        param_constraints: Dict[str, Constraint],
        *,
        validate_defaults: Container[str] | None = None,
        joint_constraints:
            Dict[ParameterConstraintContext, Callable] | None = None,
        tailor_for_dataset: Dict[str, str] | None = None,
    ):
        """
        Parameters
        ----------
        param_constraints: dict
          Mapping of parameter names to parameter constraints. On validation
          an ``EnsureParameterConstraint`` instance will be created for
          each item in this dict.
        validate_defaults: container(str), optional
          If given, this is a set of parameter names for which the default
          rule, to not validate default values, does not apply and
          default values shall be passed through a given validator.
        joint_constraints: dict, optional
          Specification of higher-order constraints considering multiple
          parameters together. See the ``joint_validation()`` method for
          details. Constraints will be processed in the order in which
          they are declared in the mapping. Earlier validators can modify
          the parameter values that are eventually passed to validators
          executed later.
        tailor_for_dataset: dict, optional
          If given, this is a mapping of a name of a parameter whose
          constraint should be tailored to a particular dataset, to a name
          of a parameter providing this dataset. The dataset-providing
          parameter constraints will be evaluated first, and the resulting
          Dataset instances are used to tailor the constraints that
          require a dataset-context. The tailoring is performed if, and
          only if, the dataset-providing parameter actually evaluated
          to a `Dataset` instance. The non-tailored constraint is used
          otherwise.
        """
        super().__init__()
        self._param_constraints = param_constraints
        self._joint_constraints = joint_constraints
        self._validate_defaults = validate_defaults or set()
        self._tailor_for_dataset = tailor_for_dataset or {}

    def joint_validation(self, params: Dict, on_error: str) -> Dict:
        """Higher-order validation considering multiple parameters at a time

        This method is called with all, individually validated, command
        parameters in keyword-argument form in the ``params`` dict argument.

        Arbitrary additional validation steps can be performed on the full
        set of parameters that may involve raising exceptions on validation
        errors, but also value transformation or replacements of individual
        parameters based on the setting of others.

        The parameter values returned by the method are passed on to the
        respective command implementation.

        The default implementation iterates over the ``joint_validators``
        specification given to the constructor, in order to perform
        any number of validations. This is a mapping of a
        ``ParameterConstraintContext`` instance to a callable implementing a
        validation for a particular parameter set.

        Example::

          _joint_validators_ = {
              ParameterConstraintContext(('p1', 'p2'), 'sum'):
                  MyValidator._check_sum,
          }

          def _checksum(self, p1, p2):
              if (p1 + p2) < 3:
                  self.raise_for(
                     dict(p1=p1, p2=p2),
                     'parameter sum is too large',
                  )

        The callable will be passed the arguments named in the
        ``ParameterConstraintContext`` as keyword arguments, using the same
        names as originally given to ``EnsureCommandParameterization``.

        Any raised ``ConstraintError`` is caught and reported together with the
        respective ``ParameterConstraintContext``. The violating value reported
        in such a ``ConstraintError`` must be a mapping of parameter name to
        value, comprising the full parameter set (i.e., keys matching the
        ``ParameterConstraintContext``).  The use of ``self.raise_for()`` is
        encouraged.

        If the callable anyhow modifies the passed arguments, it must return
        them as a kwargs-like mapping.  If nothing is modified, it is OK to
        return ``None``.

        Returns
        -------
        dict
          The returned dict must have a value for each item passed in via
          ``params``.
        on_error: {'raise-early', 'raise-at-end'}
          Flag how to handle constraint violation. By default, validation is
          stopped at the first error and an exception is raised. When an
          exhaustive validation is performed, an eventual exception contains
          information on all constraint violations.

        Raises
        ------
        ConstraintErrors
          With `on_error='raise-at-end'` an implementation can choose to
          collect more than one higher-order violation and raise them
          as a `ConstraintErrors` exception.
        """
        # if we have nothing, do nothing
        if not self._joint_constraints:
            return params

        exceptions = {}
        validated = params.copy()

        for ctx, validator in self._joint_constraints.items():
            # what the validator will produce
            res = None
            try:
                # call the validator with the parameters given in the context
                # and only with those, to make sure the context is valid
                # and not an underspecification.
                # pull the values form `validated` to be able to benefit
                # from incremental coercing done in individual checks
                res = validator(**{p: validated[p] for p in ctx.parameters})
            except ConstraintError as e:
                if not isinstance(e.value, dict) \
                        or set(ctx.parameters) != e.value.keys():  # pragma: no cover
                    raise RuntimeError(
                        'on raising a ConstraintError the joint validator '
                        f'{validator} did not report '
                        'a mapping of parameter name to (violating) value '
                        'comprising all constraint context parameters. '
                        'This is a software defect of the joint validator. '
                        'Please report!')
                exceptions[ctx] = e
                if on_error == 'raise-early':
                    raise CommandParametrizationError(exceptions)
            if res is not None:
                validated.update(**res)

        if exceptions:
            raise CommandParametrizationError(exceptions)

        return validated

    def __call__(
        self,
        kwargs,
        at_default=None,
        required=None,
        on_error='raise-early',
    ) -> Dict:
        """
        Parameters
        ----------
        kwargs: dict
          Parameter name (``str``)) to value (any) mapping of the parameter
          set.
        at_default: set or None
          Set of parameter names where the respective values in ``kwargs``
          match their respective defaults. This is used for deciding whether
          or not to process them with an associated value constraint (see the
          ``validate_defaults`` constructor argument).
        required: set or None
          Set of parameter names that are known to be required.
        on_error: {'raise-early', 'raise-at-end'}
          Flag how to handle constraint violation. By default, validation is
          stopped at the first error and an exception is raised. When an
          exhaustive validation is performed, an eventual exception contains
          information on all constraint violations. Regardless of this mode
          more than one error can be reported (in case (future) implementation
          perform independent validations in parallel).

        Raises
        ------
        CommandParametrizationError
          Raised whenever one (or more) ``ConstraintError`` exceptions are
          caught during validation. Other exception types are not caught and
          pass through.
        """
        assert on_error in ('raise-early', 'raise-at-end')

        exceptions = {}
        missing_args = tuple(a for a in (required or []) if a not in kwargs)
        if missing_args:
            exceptions[ParameterConstraintContext(missing_args)] = \
                ConstraintError(
                    self,
                    dict(zip(missing_args, [NoValue()] * len(missing_args))),
                    'missing required arguments',
                )
            if on_error == 'raise-early':
                raise CommandParametrizationError(exceptions)

        # validators to work with. make a copy of the dict to be able to tailor
        # them for this run only
        # TODO copy likely not needed
        param_constraints = self._param_constraints.copy()

        # names of parameters we need to process
        to_validate = set(kwargs)
        # check for any dataset that are required for tailoring other parameters
        ds_provider_params = set(self._tailor_for_dataset.values())
        # take these out of the set of parameters to validate, because we need
        # to process them first.
        # the approach is to simply sort them first, but otherwise apply standard
        # handling
        to_validate.difference_update(ds_provider_params)
        # strip all args provider args that have not been provided
        ds_provider_params.intersection_update(kwargs)

        validated = {}
        # process all parameters. starts with those that are needed as
        # dependencies for others.
        # this dependency-based sorting is very crude for now. it does not
        # consider possible dependencies within `ds_provider_params` at all
        for argname in chain(ds_provider_params, to_validate):
            arg = kwargs[argname]
            if at_default \
                    and argname not in self._validate_defaults \
                    and argname in at_default:
                # do not validate any parameter where the value matches the
                # default declared in the signature. Often these are just
                # 'do-nothing' settings or have special meaning that need
                # not be communicated to a user. Not validating them has
                # two consequences:
                # - the condition can simply be referred to as "default
                #   behavior" regardless of complexity
                # - a command implementation must always be able to handle
                #   its own defaults directly, and cannot delegate a
                #   default value handling to a constraint
                #
                # we must nevertheless pass any such default value through
                # to make/keep them accessible to the general result handling
                # code
                validated[argname] = arg
                continue

            # look-up validator for this parameter, if there is none use
            # NoConstraint to avoid complex conditionals in the code below
            validator = param_constraints.get(argname, NoConstraint())

            # do we need to tailor this constraint for a specific dataset?
            # only do if instructed AND the respective other parameter
            # validated to a Dataset instance. Any such parameter was sorted
            # to be validated first in this loop, so the outcome of that is
            # already available
            tailor_for = self._tailor_for_dataset.get(argname)
            if tailor_for and isinstance(validated.get(tailor_for),
                                         DatasetParameter):
                validator = validator.for_dataset(validated[tailor_for])

            try:
                validated[argname] = validator(arg)
            # we catch only ConstraintError -- only these exceptions have what
            # we need for reporting. If any validator chooses to raise
            # something else, we do not handle it here, but let it bubble up.
            # it may be an indication of something being wrong with validation
            # itself
            except ConstraintError as e:
                # standard exception type, record and proceed
                exceptions[ParameterConstraintContext((argname,))] = e
                if on_error == 'raise-early':
                    raise CommandParametrizationError(exceptions)
            except Exception as e:
                # non-standard exception type
                # we need to achieve uniform CommandParametrizationError
                # raising, so let's create a ConstraintError for this
                # exception
                e = ConstraintError(
                    validator, arg, '{__caused_by__}',
                    ctx=dict(__caused_by__=e),
                )
                exceptions[ParameterConstraintContext((argname,))] = e
                if on_error == 'raise-early':
                    raise CommandParametrizationError(exceptions)

        # do not bother with joint validation when the set of expected
        # arguments is not complete
        expected_for_joint_validation = set()
        for jv in self._joint_constraints or []:
            expected_for_joint_validation.update(jv.parameters)

        if not expected_for_joint_validation.issubset(validated):
            raise CommandParametrizationError(exceptions)

        try:
            # call (subclass) method to perform holistic, cross-parameter
            # validation of the full parameterization
            final = self.joint_validation(validated, on_error)
            # check requirements of .joint_validation(), a particular
            # implementation could be faulty, and we want to report this
            # problem in the right context
            try:
                assert final.keys() == validated.keys()
            except Exception as e:
                raise RuntimeError(
                    f"{self.__class__.__name__}.joint_validation() "
                    "did not return items for all passed parameters. "
                    "Invalid implementation.") from e
        # we catch the good stuff first. the underlying implementation is
        # providing an exception with detailed context info on possibly
        # multiple errors
        except ParametrizationErrors as e:
            # we can simply suck in the reports, the context keys do not
            # overlap, unless the provided validators want that for some
            # reason
            exceptions.update(e.errors)

        if exceptions:
            raise CommandParametrizationError(exceptions)

        return final



import pathlib
import pytest

from ..base import DatasetParameter
from ..basic import (
    EnsureInt,
    EnsureFloat,
    EnsureBool,
    EnsureStr,
    EnsureStrPrefix,
    EnsureNone,
    EnsureCallable,
    EnsureChoice,
    EnsureHashAlgorithm,
    EnsureKeyChoice,
    EnsureRange,
    EnsurePath,
    EnsureValue,
    NoConstraint,
)
from ..exceptions import ConstraintError

from ..utils import _type_str


def test_noconstraint():
    c = NoConstraint()
    assert c(5) == 5
    assert c.short_description() == ''


def test_int():
    c = EnsureInt()
    # this should always work
    assert c(7) == 7
    assert c(7.0) == 7
    assert c('7') == 7
    # no automatic inspection of iterables, should use EnsureIterableOf
    with pytest.raises(ConstraintError):
        c([7, 3])
    # this should always fail
    with pytest.raises(ValueError):
        c('fail')
    # this will also fail
    with pytest.raises(ValueError):
        c('17.0')
    assert c.short_description() == 'int'


def test_float():
    c = EnsureFloat()
    # this should always work
    assert c(7.0) == 7.0
    assert c(7) == 7.0
    assert c('7') == 7.0
    # no automatic inspection of iterables, should use EnsureIterableOf
    with pytest.raises(ConstraintError):
        c([7.0, '3.0'])
    # this should always fail
    with pytest.raises(ValueError):
        c('fail')


def test_bool():
    c = EnsureBool()
    # this should always work
    assert c(True) is True
    assert c(False) is False
    # all that results in True
    assert c('True') is True
    assert c('true') is True
    assert c('1') is True
    assert c('yes') is True
    assert c('on') is True
    assert c('enable') is True
    # all that results in False
    assert c('false') is False
    assert c('False') is False
    assert c('0') is False
    assert c('no') is False
    assert c('off') is False
    assert c('disable') is False
    # this should always fail
    with pytest.raises(ValueError):
        c(0)
    with pytest.raises(ValueError):
        c(1)


def test_str():
    c = EnsureStr()
    # this should always work
    assert c('hello') == 'hello'
    assert c('7.0') == '7.0'
    # this should always fail
    with pytest.raises(ValueError):
        c(['ab'])
    with pytest.raises(ValueError):
        c(['a', 'b'])
    with pytest.raises(ValueError):
        c(('a', 'b'))
    # no automatic conversion attempted
    with pytest.raises(ValueError):
        c(7.0)
    assert c.short_description() == 'str'


def test_str_min_len():
    c = EnsureStr(min_len=1)
    assert c('hello') == 'hello'
    assert c('h') == 'h'
    with pytest.raises(ValueError):
        c('')

    c = EnsureStr(min_len=2)
    assert c('hello') == 'hello'
    with pytest.raises(ValueError):
        c('h')


def test_EnsureStr_match():
    # alphanum plus _ and ., non-empty
    pattern = '[a-zA-Z0-9-.]+'
    constraint = EnsureStr(match=pattern)

    # reports the pattern in the description
    for m in (constraint.short_description, constraint.long_description):
        assert pattern in m()

    # must work
    assert constraint('a0F-2.') == 'a0F-2.'

    for v in ('', '123_abc'):
        with pytest.raises(ValueError):
            assert constraint('')


def test_EnsureStrPrefix():
    c = EnsureStrPrefix('some-')
    c('some-mess') == 'some-mess'
    with pytest.raises(ValueError):
        c('mess')
    assert c.short_description() == 'some-...'
    assert c.long_description() == "value must start with 'some-'"


def test_EnsureValue():
    c = EnsureValue(5)
    assert c.short_description() == '5'
    # this should always work
    assert c(5) == 5
    # type mismatch
    with pytest.raises(ValueError):
        c('5')
    # value mismatches
    with pytest.raises(ValueError):
        c('None')
    with pytest.raises(ValueError):
        c([])


# special case of EnsureValue
def test_none():
    c = EnsureNone()
    assert c.short_description() == 'None'
    # this should always work
    assert c(None) is None
    # this should always fail
    with pytest.raises(ValueError):
        c('None')
    with pytest.raises(ValueError):
        c([])


def test_callable():
    c = EnsureCallable()
    assert c.short_description() == 'callable'
    assert c.long_description() == 'value must be a callable'
    # this should always work
    assert c(range) == range
    with pytest.raises(ValueError):
        c('range')


def test_choice():
    c = EnsureChoice('choice1', 'choice2', None)
    descr = c.long_description()
    for i in ('choice1', 'choice2', 'CMD', 'PY'):
        assert i in descr
    # short is a "set" or repr()s
    assert c.short_description() == "{'choice1', 'choice2', None}"
    assert str(c) == "one of {'choice1', 'choice2', None}"
    # this should always work
    assert c('choice1') == 'choice1'
    assert c(None) is None
    # this should always fail
    with pytest.raises(ValueError):
        c('fail')
    with pytest.raises(ValueError):
        c('None')


def test_keychoice():
    c = EnsureKeyChoice(key='some', values=('choice1', 'choice2', None))
    descr = c.long_description()
    for i in ('some', 'choice1', 'choice2'):
        assert i in descr
    assert c.short_description() == "some:{'choice1', 'choice2', None}"
    assert c({'some': 'choice1'}) == {'some': 'choice1'}
    assert c({'some': None}) == {'some': None}
    assert c({'some': None, 'ign': 'ore'}) == {'some': None, 'ign': 'ore'}
    with pytest.raises(ValueError):
        c('fail')
    with pytest.raises(ValueError):
        c('None')
    with pytest.raises(ValueError):
        c({'nope': 'None'})
    with pytest.raises(ValueError):
        c({'some': 'None'})
    with pytest.raises(ValueError):
        c({'some': ('a', 'b')})


def test_range():
    with pytest.raises(ValueError):
        EnsureRange(min=None, max=None)
    c = EnsureRange(max=7)
    assert c.short_description() == 'not greater than 7'
    c = EnsureRange(min=3, max=7)
    # this should always work
    assert c(3.0) == 3.0

    # this should always fail
    with pytest.raises(ValueError):
        c(2.9999999)
    with pytest.raises(ValueError):
        c(77)
    with pytest.raises(TypeError):
        c('fail')
    with pytest.raises(TypeError):
        c((3, 4))
    # since no type checks are performed
    with pytest.raises(TypeError):
        c('7')

    # Range doesn't have to be numeric
    c = EnsureRange(min="e", max="qqq")
    assert c.short_description() == "in range from 'e' to 'qqq'"
    assert c('e') == 'e'
    assert c('fa') == 'fa'
    assert c('qq') == 'qq'
    with pytest.raises(ValueError):
        c('a')
    with pytest.raises(ValueError):
        c('qqqa')


def test_type_str():
    assert _type_str((str,)) == 'str'
    assert _type_str(str) == 'str'


def test_EnsurePath(tmp_path):
    target = pathlib.Path(tmp_path)

    assert EnsurePath()(tmp_path) == target
    assert EnsurePath(lexists=True)(tmp_path) == target
    with pytest.raises(ValueError):
        EnsurePath(lexists=False)(tmp_path)
    with pytest.raises(ValueError):
        EnsurePath(lexists=True)(tmp_path / 'nothere')
    assert EnsurePath(is_format='absolute')(tmp_path) == target
    with pytest.raises(ValueError):
        EnsurePath(is_format='relative')(tmp_path)
    with pytest.raises(ValueError):
        EnsurePath(is_format='absolute')(tmp_path.name)
    from stat import S_ISDIR, S_ISREG
    assert EnsurePath(is_mode=S_ISDIR)(tmp_path) == target
    with pytest.raises(ValueError):
        EnsurePath(is_mode=S_ISREG)(tmp_path)
    # give particular path type
    assert EnsurePath(path_type=pathlib.PurePath
        )(tmp_path) == pathlib.PurePath(tmp_path)
    # not everything is possible, this is known and OK
    with pytest.raises(AttributeError):
        EnsurePath(
            path_type=pathlib.PurePath,
            is_mode=S_ISREG,
        )(tmp_path)
    assert EnsurePath().short_description() == 'path'
    assert EnsurePath(is_format='absolute').short_description() == 'absolute path'
    # default comparison mode is parent-or-same-as
    c = EnsurePath(ref=target)
    assert c(target) == target
    assert c(target / 'some') == target / 'some'
    with pytest.raises(ValueError):
        assert c(target.parent)
    c = EnsurePath(ref=target, ref_is='parent-of')
    assert c(target / 'some') == target / 'some'
    with pytest.raises(ValueError):
        assert c(target)
    assert c.short_description() == f'path that is parent-of {target}'
    with pytest.raises(AssertionError):
        c = EnsurePath(ref=target, ref_is='stupid')


def test_EnsurePath_fordataset(existing_dataset):
    P = pathlib.Path
    ds = existing_dataset
    # standard: relative in, relative out
    c = EnsurePath()
    assert c('relpath') == P('relpath')
    # tailor constraint for our dataset
    # (this is what would be done by EnsureCommandParameterization
    # 1. dataset given as a path -- resolve against CWD
    #    output is always absolute
    tc = c.for_dataset(DatasetParameter(None, ds))
    assert tc('relpath') == (P.cwd() / 'relpath')
    # 2. dataset is given as a dataset object
    tc = c.for_dataset(DatasetParameter(ds, ds))
    assert tc('relpath') == (ds.pathobj / 'relpath')


def test_EnsureHashAlgorithm():
    c = EnsureHashAlgorithm()
    # simple cases that should pass
    hashes = [
        'sha3_256', 'shake_256', 'sha3_384', 'md5', 'shake_128', 'sha384',
        'sha3_224', 'blake2s', 'sha1', 'blake2b', 'sha224', 'sha512', 'sha256',
        'sha3_512'
    ]
    for hash in hashes:
        c(hash)
    # a few bogus ones:
    bad_hashes = [
        'md17', 'McGyver', 'sha2', 'bogus'
    ]
    for baddie in bad_hashes:
        with pytest.raises(ConstraintError):
            c(baddie)

    # check messaging
    for i in ('md5', 'shake_256', 'sha3_512'):
        assert i in c.short_description()
        assert i in c.long_description()



"""Context manager to communicate with a subprocess using iterables

.. deprecated:: 1.6
   This module is deprecated. It has been migrated to the `datasalad library
   <https://pypi.org/project/datasalad>`__. Imports should be adjusted to
   ``datasalad.iterable_subprocess``.

This offers a higher level interface to subprocesses than Python's built-in
subprocess module, and is particularly helpful when data won't fit in memory
and has to be streamed.

This also allows an external subprocess to be naturally placed in a chain of
iterables as part of a data processing pipeline.

This code has been taken from https://pypi.org/project/iterable-subprocess/
and was subsequently adjusted for cross-platform compatibility and performance,
as well as tighter integration with DataLad.

The original code was made available under the terms of the MIT License,
and was written by Michal Charemza.
"""

__all__ = ['iterable_subprocess']

import warnings

from datasalad.iterable_subprocess import iterable_subprocess

warnings.warn(
    '`datalad_next.iterable_subprocess` has been migrated to the datasalad '
    'library, adjust imports to `datasalad.iterable_subprocess`',
    DeprecationWarning,
    stacklevel=1,
)



"""Type ENUMs"""

from enum import Enum


class ArchiveType(Enum):
    """Enumeration of archive types

    Each one should have an associated ArchiveOperations handler.
    """
    # TODO the values could also be handler classes ...
    tar = 'tar'
    zip = 'zip'



from webdav3.client import Client as DAVClient


def test_serve_webdav_fixture(webdav_credential, webdav_server):
    webdav_cfg = dict(
        webdav_hostname=webdav_server.url,
        webdav_login=webdav_credential['user'],
        webdav_password=webdav_credential['secret'],
        webdav_root='/',
    )
    webdav = DAVClient(webdav_cfg)
    # plain use should work without error
    webdav.list()
    (webdav_server.path / 'probe').touch()
    assert 'probe' in webdav.list()



"""Iterators for particular types of collections

Most importantly this includes different collections (or containers) for files,
such as a file system directory, or an archive (also see the
``ls_file_collection`` command). However, this module is not per-se limited
to file collections.

Most, if not all, implementation come in the form of a function that takes
a collection identifier or a collection location (e.g., a file system path),
and possibly some additional options. When called, an iterator is returned
that produces collection items in the form of data class instances of
a given type. The particular type can be different across different
collections.


.. currentmodule:: datalad_next.iter_collections
.. autosummary::
   :toctree: generated

   iter_annexworktree
   iter_dir
   iter_gitdiff
   iter_gitstatus
   iter_gittree
   iter_gitworktree
   iter_submodules
   iter_tar
   iter_zip
   TarfileItem
   ZipfileItem
   FileSystemItem
   FileSystemItemType
   GitTreeItemType
   GitWorktreeItem
   GitWorktreeFileSystemItem
   GitDiffItem
   GitDiffStatus
   GitContainerModificationType
"""

from .tarfile import (
    # TODO move to datalad_next.types?
    TarfileItem,
    iter_tar,
)
from .zipfile import (
    # TODO move to datalad_next.types?
    ZipfileItem,
    iter_zip,
)
# TODO move to datalad_next.types?
from .utils import (
    # TODO move to datalad_next.types?
    FileSystemItemType,
    # TODO move to datalad_next.types?
    FileSystemItem,
    compute_multihash_from_fp,
)
from .directory import iter_dir
from .gittree import (
    # TODO move to datalad_next.types?
    GitTreeItemType,
    iter_gittree,
)
from .gitworktree import (
    # TODO move to datalad_next.types?
    GitWorktreeItem,
    # TODO move to datalad_next.types?
    GitWorktreeFileSystemItem,
    iter_gitworktree,
    iter_submodules,
)
from .annexworktree import (
    iter_annexworktree,
)
from .gitdiff import (
    # TODO move to datalad_next.types?
    GitDiffItem,
    # TODO move to datalad_next.types?
    GitDiffStatus,
    # TODO move to datalad_next.types?
    GitContainerModificationType,
    iter_gitdiff,
)
from .gitstatus import (
    iter_gitstatus,
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Credential management and query"""

__docformat__ = 'restructuredtext'

import json
import logging
from typing import Dict

from datalad import (
    cfg as dlcfg,
)
from datalad_next.credman import (
    CredentialManager,
    verify_property_names,
)
from datalad_next.commands import (
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    ParameterConstraintContext,
    build_doc,
    eval_results,
    generic_result_renderer,
    get_status_dict,
)
from datalad_next.exceptions import CapturedException
from datalad_next.datasets import datasetmethod
from datalad_next.constraints import (
    EnsureChoice,
    EnsureDataset,
    EnsureNone,
    EnsureStr,
)
from datalad_next.utils import ParamDictator


lgr = logging.getLogger('datalad.local.credentials')

credential_actions = ('query', 'get', 'set', 'remove')


class CredentialsParamValidator(EnsureCommandParameterization):
    def __init__(self):
        super().__init__(
            param_constraints=dict(
                action=EnsureChoice(*credential_actions),
                dataset=EnsureDataset(
                    # if given, it must also exist as a source for
                    # configuration items and/or credentials
                    installed=True,
                    purpose='manage credentials',
                ),
                name=EnsureStr(),
                prompt=EnsureStr(),
            ),
            # order in joint_constraints is relevant!
            joint_constraints={
                ParameterConstraintContext(('action', 'name', 'spec'),
                                           'CLI normalization'):
                self._normalize_cli_params,
                ParameterConstraintContext(('spec',),
                                           'credential spec normalization'):
                self._normalize_spec,
                # check parameter requirements for particular actions
                ParameterConstraintContext(('action', 'name'),
                                           'remove-action requirements'):
                self._check_remove_requirements,
                ParameterConstraintContext(('action', 'name', 'spec'),
                                           'get-action requirements'):
                self._check_get_requirements,
            },
        )

    def _normalize_cli_params(self, action, name, spec):
        if action in ('get', 'set', 'remove') and not name and spec \
                and isinstance(spec, list):
            # spec came in like from the CLI (but doesn't have to be from
            # there) and we have no name set
            if spec[0][0] != ':' and '=' not in spec[0]:
                name = spec[0]
                spec = spec[1:]
        return dict(action=action, name=name, spec=spec)

    def _normalize_spec(self, spec):
        # `spec` could be many things, make uniform dict
        return dict(spec=normalize_specs(spec))

    def _check_remove_requirements(self, action, name):
        if action == 'remove' and not name:
            self.raise_for(
                dict(action=action, name=name),
                'no credential name provided',
            )

    def _check_get_requirements(self, action, name, spec):
        if action == 'get' and not name and not spec:
            self.raise_for(
                dict(action=action, name=name, spec=spec),
                'no name or credential property specified',
            )


@build_doc
class Credentials(ValidatedInterface):
    """Credential management and query

    This command enables inspection and manipulation of credentials used
    throughout DataLad.

    The command provides four basic actions:


    QUERY

    When executed without any property specification, all known credentials
    with all their properties will be yielded. Please note that this may not
    include credentials that only comprise of a secret and no other properties,
    or legacy credentials for which no trace in the configuration can be found.
    Therefore, the query results are not guaranteed to contain all credentials
    ever configured by DataLad.

    When additional property/value pairs are specified, only credentials that
    have matching values for all given properties will be reported. This can be
    used, for example, to discover all suitable credentials for a specific
    "realm", if credentials were annotated with such information.


    SET

    This is the companion to 'get', and can be used to store properties and
    secret of a credential. Importantly, and in contrast to a 'get' operation,
    given properties with no values indicate a removal request. Any matching
    properties on record will be removed. If a credential is to be stored for
    which no secret is on record yet, an interactive session will prompt a user
    for a manual secret entry.

    Only changed properties will be contained in the result record.

    The appearance of the interactive secret entry can be configured with
    the two settings `datalad.credentials.repeat-secret-entry` and
    `datalad.credentials.hidden-secret-entry`.


    REMOVE

    This action will remove any secret and properties associated with a
    credential identified by its name.


    GET (plumbing operation)

    This is a *read-only* action that will never store (updates of) credential
    properties or secrets. Given properties will amend/overwrite those already
    on record.  When properties with no value are given, and also no value for
    the respective properties is on record yet, their value will be requested
    interactively, if a ``prompt||--prompt`` text was provided too. This can be
    used to ensure a complete credential record, comprising any number of
    properties.


    Details on credentials

    A credential comprises any number of properties, plus exactly one secret.
    There are no constraints on the format or property values or the secret,
    as long as they are encoded as a string.

    Credential properties are normally stored as configuration settings in a
    user's configuration ('global' scope) using the naming scheme:

      `datalad.credential.<name>.<property>`

    Therefore both credential name and credential property name must be
    syntax-compliant with Git configuration items. For property names this
    means only alphanumeric characters and dashes. For credential names
    virtually no naming restrictions exist (only null-byte and newline are
    forbidden). However, when naming credentials it is recommended to use
    simple names in order to enable convenient one-off credential overrides
    by specifying DataLad configuration items via their environment variable
    counterparts (see the documentation of the ``configuration`` command
    for details. In short, avoid underscores and special characters other than
    '.' and '-'.

    While there are no constraints on the number and nature of credential
    properties, a few particular properties are recognized on used for
    particular purposes:

    - 'secret': always refers to the single secret of a credential
    - 'type': identifies the type of a credential. With each standard type,
      a list of mandatory properties is associated (see below)
    - 'last-used': is an ISO 8601 format time stamp that indicated the
      last (successful) usage of a credential

    Standard credential types and properties

    The following standard credential types are recognized, and their
    mandatory field with their standard names will be automatically
    included in a 'get' report.

    - 'user_password': with properties 'user', and the password as secret
    - 'token': only comprising the token as secret
    - 'aws-s3': with properties 'key-id', 'session', 'expiration', and the
      secret_id as the credential secret

    Legacy support

    DataLad credentials not configured via this command may not be fully
    discoverable (i.e., including all their properties). Discovery of
    such legacy credentials can be assisted by specifying a dedicated
    'type' property.
    """
    result_renderer = 'tailored'

    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""specify a dataset whose configuration to inspect
            rather than the global (user) settings"""),
        action=Parameter(
            args=("action",),
            nargs='?',
            doc="""which action to perform"""),
        name=Parameter(
            # exclude from CLI
            args=tuple(),
            doc="""name of a credential to set, get, or remove."""),
        spec=Parameter(
            args=("spec",),
            doc="""specification of[CMD: a credential name and CMD]
            credential properties. Properties are[CMD:  either CMD] given as
            name/value pairs[CMD:  or as a property name prefixed
            by a colon CMD].
            Properties [CMD: prefixed with a colon CMD][PY: with a `None` value PY] 
            indicate a property to be deleted (action 'set'), or a
            property to be entered interactively, when no value is set
            yet, and a prompt text is given (action 'get').
            All property names are case-insensitive, must start with
            a letter or a digit, and may only contain '-' apart from
            these characters.
            [PY: Property specifications should be given a as dictionary,
            e.g., spec={'type': 'user_password'}.
            However, a CLI-like list of string arguments is also
            supported, e.g., spec=['type=user_password'] PY]""",
            nargs='*',
            metavar='[name] [:]property[=value]'),
        prompt=Parameter(
            args=("--prompt",),
            doc="""message to display when entry of missing credential
            properties is required for action 'get'. This can be used
            to present information on the nature of a credential and
            for instructions on how to obtain a credential"""),
    )

    _examples_ = [
        dict(text="Report all discoverable credentials",
             code_py="credentials()",
             code_cmd="datalad credentials"),
        dict(
            text="Set a new credential mycred & input its secret interactively",
            code_py="credentials('set', name='mycred')",
            code_cmd="datalad credentials set mycred"),
        dict(text="Remove a credential's type property",
             code_py="credentials('set', name='mycred', spec={'type': None})",
             code_cmd="datalad credentials set mycred :type"),
        dict(text="Get all information on a specific credential in a structured record",
             code_py="credentials('get', name='mycred')",
             code_cmd="datalad -f json credentials get mycred"),
        dict(text="Upgrade a legacy credential by annotating it with a 'type' property",
             code_py="credentials('set', name='legacycred', spec={'type': 'user_password')",
             code_cmd="datalad credentials set legacycred type=user_password"),
        dict(text="Set a new credential of type user_password, with a given user property, "
                  "and input its secret interactively",
             code_py="credentials('set', name='mycred', spec={'type': 'user_password', 'user': '<username>'})",
             code_cmd="datalad credentials set mycred type=user_password user=<username>"),
        dict(text="Obtain a (possibly yet undefined) credential with a minimum set of "
                  "properties. All missing properties and secret will be "
                  "prompted for, no information will be stored! "
                  "This is mostly useful for ensuring availability of an "
                  "appropriate credential in an application context",
             code_py="credentials('get', prompt='Can I haz info plz?', name='newcred', spec={'newproperty': None})",
             code_cmd="datalad credentials --prompt 'can I haz info plz?' get newcred :newproperty"),
    ]

    _validator_ = CredentialsParamValidator()

    @staticmethod
    @datasetmethod(name='credentials')
    @eval_results
    def __call__(action='query', spec=None, *, name=None, prompt=None,
                 dataset=None):
        # which config manager to use: global or from dataset
        # It makes no sense to work with a non-existing dataset's config,
        # due to https://github.com/datalad/datalad/issues/7299
        # so the `dataset` validator must not run for the default value
        # ``None``
        cfg = dataset.ds.config if dataset else dlcfg

        credman = CredentialManager(cfg)

        if action == 'set':
            try:
                updated = credman.set(name, **spec)
            except Exception as e:
                yield get_status_dict(
                    action='credentials',
                    status='error',
                    name=name,
                    message='could not set credential properties',
                    exception=CapturedException(e),
                )
                return
            # pull name out of report, if entered manually
            if not name and updated is not None:
                name = updated.pop('name', None)
            yield get_status_dict(
                action='credentials',
                status='notneeded' if updated is None else 'ok',
                name=name,
                **_prefix_result_keys(updated if updated else spec),
            )
        elif action == 'get':
            cred = credman.get(name=name, _prompt=prompt, **spec)
            if not cred:
                yield get_status_dict(
                    action='credentials',
                    status='error',
                    name=name,
                    message='credential not found',
                )
            else:
                yield get_status_dict(
                    action='credentials',
                    status='ok',
                    name=name,
                    **_prefix_result_keys(cred),
                )
        elif action == 'remove':
            try:
                removed = credman.remove(name, type_hint=spec.get('type'))
            except Exception as e:
                yield get_status_dict(
                    action='credentials',
                    status='error',
                    name=name,
                    message='could not remove credential properties',
                    exception=CapturedException(e),
                )
                return
            yield get_status_dict(
                action='credentials',
                status='ok' if removed else 'notneeded',
                name=name,
            )
        elif action == 'query':
            for name, cred in credman.query_(**spec):
                yield get_status_dict(
                    action='credentials',
                    status='ok',
                    name=name,
                    type='credential',
                    **_prefix_result_keys(cred),
                )
        else:
            raise RuntimeError('Impossible state reached')  # pragma: no cover

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        # we only handle our own stuff in a custom fashion, the rest is generic
        if res['action'] != 'credentials':
            generic_result_renderer(res)
            return
        # must make a copy, because we modify the record in-place
        # https://github.com/datalad/datalad/issues/6560
        res = res.copy()
        # the idea here is to twist the result records such that the generic
        # renderer can be used
        if 'name' in res:
            res['action'] = res['name']
        res['status'] = '{} {}'.format(
            res.get('cred_type', 'secret'),
            '✓' if res.get('cred_secret') else '✗',
        )

        if res.pop('from_backend', None) == 'legacy':
            res['type'] = 'legacy-credential'
        if 'message' not in res:
            # give the names of all properties
            # but avoid duplicating the type, hide the prefix,
            # add removal marker for vanished properties
            res['message'] = ','.join(
                '{}{}{}{}'.format(
                    f':{p[5:]}' if res[p] is None else p[5:],
                    '' if res[p] is None else '=',
                    '' if res[p] is None else res[p][:25],
                    '…' if res[p] and len(res[p]) > 25 else '',
                )
                for p in sorted(res)
                if p.startswith('cred_') and p not in ('cred_type', 'cred_secret'))
        generic_result_renderer(res)


def normalize_specs(specs):
    """Normalize all supported `spec` argument values for `credentials`

    Parameter
    ---------
    specs: JSON-formatted str or list

    Returns
    -------
    dict
        Keys are the names of any property (with removal markers stripped),
        and values are `None` whenever property removal is desired, and
        not `None` for any value to be stored.

    Raises
    ------
    ValueError
      For missing values, missing removal markers, and invalid JSON input
    """
    if not specs:
        return {}
    elif isinstance(specs, str):
        try:
            specs = json.loads(specs)
        except json.JSONDecodeError as e:
            raise ValueError('Invalid JSON input') from e
    if isinstance(specs, list):
        # convert property assignment list
        specs = [
            (str(s[0]), str(s[1]))
            if isinstance(s, tuple) else
            (str(s),)
            if '=' not in s else
            (tuple(s.split('=', 1)))
            for s in specs
        ]
    if isinstance(specs, list):
        missing = [
            i for i in specs
            if (len(i) == 1 and i[0][0] != ":") or (
                len(i) > 1 and (i[0][0] == ':' and i[1] is not None))
        ]
    else:
        missing = [
            k for k, v in specs.items()
            if k[0] == ":" and v is not None
        ]
    if missing:
        raise ValueError(
            f'Value or unset flag ":" missing for property {missing!r}')
    if isinstance(specs, list):
        # expand absent values in tuples to ease conversion to dict below
        specs = [(i[0], i[1] if len(i) > 1 else None) for i in specs]
    # apply "unset marker"
    specs = {
        # this stuff goes to git-config, is therefore case-insensitive
        # and we should normalize right away
        (k[1:] if k[0] == ':' else k).lower():
        None if k[0] == ':' else v
        for k, v in (specs.items() if isinstance(specs, dict) else specs)
    }
    verify_property_names(specs)
    return specs


def _prefix_result_keys(props):
    return {
        f'cred_{k}' if not k.startswith('_') else k[1:]: v
        for k, v in props.items()
    }



from __future__ import annotations

from typing import cast

import pytest

from datasalad.iterable_subprocess.iterable_subprocess import OutputFrom
from ..response_generators import (
    FixedLengthResponseGeneratorPosix,
    VariableLengthResponseGeneratorPosix,
    lgr as response_generator_lgr
)
from ..operations.posix import DownloadResponseGeneratorPosix
from ..operations.common import lgr as posix_common_lgr


class DummyOutputFrom(OutputFrom):
    def __init__(self,
                 iterable: list[bytes] | None = None,
                 ) -> None:
        super().__init__(None, None)
        self.iterable = iterable

    def send(self, _):
        if self.iterable:
            return self.iterable.pop(0)
        raise StopIteration


def test_unknown_state_detection_in_variable():
    # Check that the response generator detects an unknown internal state.
    # Since different
    response_generator = VariableLengthResponseGeneratorPosix(
        cast(OutputFrom, DummyOutputFrom())
    )
    response_generator.state = '<-no-such-state->'
    with pytest.raises(RuntimeError):
        response_generator.send(b'')


def test_unknown_state_detection():
    # Check that the response generator detects an unknown internal state.
    # Since different
    response_generators = [
        VariableLengthResponseGeneratorPosix(
            cast(OutputFrom, DummyOutputFrom())
        ),
        FixedLengthResponseGeneratorPosix(
            cast(OutputFrom, DummyOutputFrom()),
            100,
        ),
        DownloadResponseGeneratorPosix(
            cast(OutputFrom, DummyOutputFrom())
        ),
    ]
    for response_generator in response_generators:
        response_generator.state = '<-no-such-state->'
        with pytest.raises(RuntimeError):
            response_generator.send(b'')


def test_trailing_content_detection_in_variable(monkeypatch):
    # Check that the response generator detects a trailing newline.
    input_list = []
    warning_list = []
    response_generator = VariableLengthResponseGeneratorPosix(
        cast(OutputFrom, DummyOutputFrom(input_list))
    )
    input_list.extend([
        b'123\n',
        response_generator.stream_marker,
        b'0\nEXTRA-CONTENT\n',
    ])
    monkeypatch.setattr(
        response_generator_lgr,
        'warning',
        lambda *args, **kwargs: warning_list.append((args, kwargs))
    )
    assert tuple(response_generator) == (b'123\n',)
    assert warning_list == [(
        ('unexpected output after return code: %s', "b'EXTRA-CONTENT\\n'"),
        {}
    )]


def test_trailing_content_detection_in_fixed(monkeypatch):
    # Check that the response generator detects a trailing newline.
    input_list = [b'1230\nEXTRA-CONTENT\n']
    warning_list = []
    response_generator = FixedLengthResponseGeneratorPosix(
        cast(OutputFrom, DummyOutputFrom(input_list)),
        3,
    )
    monkeypatch.setattr(
        response_generator_lgr,
        'warning',
        lambda *args, **kwargs: warning_list.append((args, kwargs))
    )
    assert tuple(response_generator) == (b'123',)
    assert warning_list == [(
        ('unexpected output after return code: %s', "b'EXTRA-CONTENT\\n'"),
        {}
    )]


def test_trailing_content_detection_in_download(monkeypatch):
    # Check that the response generator detects a trailing newline.
    input_list = [b'3\n1230\nEXTRA-CONTENT\n']
    warning_list = []
    response_generator = DownloadResponseGeneratorPosix(
        cast(OutputFrom, DummyOutputFrom(input_list)),
    )
    monkeypatch.setattr(
        posix_common_lgr,
        'warning',
        lambda *args, **kwargs: warning_list.append((args, kwargs))
    )
    assert tuple(response_generator) == (b'123',)
    assert warning_list == [(
        ('unexpected output after return code: %s', "b'EXTRA-CONTENT\\n'"),
        {}
    )]



from pathlib import (
    PurePath,
    PurePosixPath,
)

import pytest

from datalad_next.utils import (
    check_symlink_capability,
    rmtree,
)

from ..gittree import (
    GitTreeItemType,
)
from ..gitworktree import (
    GitWorktreeItem,
    GitWorktreeFileSystemItem,
    iter_gitworktree,
    iter_submodules,
)


def test_iter_gitworktree(existing_dataset):
    ds = existing_dataset

    (ds.pathobj / 'emptydir').mkdir()
    untracked = ds.pathobj / 'subdir' / 'untracked'
    untracked.parent.mkdir()
    untracked.write_text('untracked')

    tracked_items = list(iter_gitworktree(ds.pathobj, untracked=None))
    # without untracked's and no link resolution this is plain and fast
    assert all(
        isinstance(i, GitWorktreeItem) and i.gitsha and i.gittype
        for i in tracked_items
    )

    all_items = list(iter_gitworktree(ds.pathobj, untracked='all'))
    # empty-dir is not reported, only untracked files
    assert len(all_items) == len(tracked_items) + 1
    assert any(
        i.name == PurePath('subdir', 'untracked')
        and i.gitsha is None and i.gittype is None
        for i in all_items
    )
    # same again, but with a different untracked reporting
    all_items = list(iter_gitworktree(ds.pathobj, untracked='whole-dir'))
    # emptydir is reported too
    assert len(all_items) == len(tracked_items) + 2
    assert any(
        i.name == PurePath('subdir')
        and i.gitsha is None and i.gittype is None
        for i in all_items
    )
    # and again for the last variant
    all_items = list(iter_gitworktree(ds.pathobj, untracked='no-empty-dir'))
    # and again no emptydir
    assert len(all_items) == len(tracked_items) + 1
    assert any(
        i.name == PurePath('subdir')
        and i.gitsha is None and i.gittype is None
        for i in all_items
    )

    # if we ask for file objects or link_targets, we get a different item type,
    # but including the same
    for kwargs in (
            dict(link_target=True, fp=False, untracked=None),
            dict(link_target=False, fp=True, untracked=None),
            dict(link_target=True, fp=True, untracked=None),
    ):
        assert all(
            isinstance(i, GitWorktreeFileSystemItem) and i.gitsha and i.gittype
            for i in iter_gitworktree(ds.pathobj, **kwargs)
        )

    # check that file pointers work for tracked and untracked content
    checked_tracked = False
    checked_untracked = False
    for item in iter_gitworktree(ds.pathobj, fp=True):
        if item.name == PurePath('.datalad', 'config'):
            assert ds.id in (ds.pathobj / item.name).read_text()
            checked_tracked = True
        elif item.name == PurePath('subdir', 'untracked'):
            assert 'untracked' == (ds.pathobj / item.name).read_text()
            checked_untracked = True
    assert checked_tracked
    assert checked_untracked


def test_name_starting_with_tab(existing_dataset, no_result_rendering):
    ds = existing_dataset
    if ds.repo.is_crippled_fs():
        pytest.skip("not applicable on crippled filesystems")
    tabbed_file_name = "\ttab.txt"
    tabbed_name = ds.pathobj / tabbed_file_name
    tabbed_name.write_text('name of this file starts with a tab')
    ds.save()

    iter_names = [item.name for item in iter_gitworktree(ds.pathobj)]
    assert PurePosixPath(tabbed_file_name) in iter_names


def test_iter_gitworktree_recursive(existing_dataset, no_result_rendering):
    # actually, this tests non-recursive, because within-repo
    # recursion is the default.
    # later, we might also test subdataset recursion here
    ds = existing_dataset
    # some tracked content
    tracked1 = ds.pathobj / 'tracked1'
    tracked2 = ds.pathobj / 'subdir' / 'tracked2'
    tracked3 = ds.pathobj / 'subdir' / 'tracked3'
    for p in (tracked1, tracked2, tracked3):
        p.parent.mkdir(exist_ok=True)
        p.write_text(p.name)
    ds.save()

    # an "invisible" directory (no content)
    (ds.pathobj / 'emptydir').mkdir()
    # untracked file in subdir
    untracked = ds.pathobj / 'subdir_u' / 'untracked'
    untracked.parent.mkdir()
    untracked.write_text('untracked')

    # matches git report with untracked=all
    all_content = set((
        PurePath('.datalad'),
        PurePath('subdir'),
        PurePath('.gitattributes'),
        PurePath('subdir_u'),
        PurePath('tracked1'),
    ))
    # without any recursion, we see all top-level content, except for
    # the empty directory with no content
    all_items = list(iter_gitworktree(ds.pathobj, recursive='no'))
    assert set(i.name for i in all_items) == all_content

    # no we test a query that gooey would want to make,
    # give me all content in a single directory, and also include any
    # untracked files and even untracked/empty directories
    all_items = list(iter_gitworktree(ds.pathobj, recursive='no',
                                      untracked='whole-dir'))
    assert set(i.name for i in all_items) == \
        all_content.union((PurePath('emptydir'),))


def test_iter_gitworktree_empty(existing_dataset, no_result_rendering):
    ds = existing_dataset
    rmtree(ds.pathobj / '.datalad')
    (ds.pathobj / '.gitattributes').unlink()
    ds.save()
    assert len(ds.status()) == 0
    all_items = list(iter_gitworktree(ds.pathobj))
    assert len(all_items) == 0


def test_iter_gitworktree_deadsymlinks(existing_dataset, no_result_rendering):
    ds = existing_dataset
    dpath = ds.pathobj / 'subdir'
    dpath.mkdir()
    fpath = dpath / 'file1'
    test_content = 'content'
    fpath.write_text(test_content)
    ds.save()
    ds.drop(fpath, reckless='availability')
    try:
        # if there is a file we can open, it should not have the content
        # (annex pointer file)
        assert fpath.read_text() != test_content
    except FileNotFoundError:
        # with dead symlinks, we end up here and that is normal
        pass
    # next one must not crash
    all_items = list(iter_gitworktree(dpath))
    # we get our "dead symlink" -- but depending on the p[latform
    # it may take a different form, hence not checking for type
    assert len(all_items) == 1
    assert all_items[0].name == PurePath('file1')


def prep_fp_tester(ds):
    # we expect to process an exact number of files below
    # 3 annexed files, 1 untracked, 1 in git,
    # and possibly 1 symlink in git, 1 symlink untracked
    # we count them up on creation, and then down on test
    fcount = 0

    content_tmpl = 'content: #ö file_{}'
    for i in ('annex1', 'annex2', 'annex3'):
        (ds.pathobj / f'file_{i}').write_text(
            content_tmpl.format(i), encoding='utf-8')
        fcount += 1
    ds.save()
    ds.drop(
        ds.pathobj / 'file_annex1',
        reckless='availability',
    )
    # and also add a file to git directly and a have one untracked too
    for i in ('untracked', 'ingit', 'deleted'):
        (ds.pathobj / f'file_{i}').write_text(
            content_tmpl.format(i), encoding='utf-8')
        fcount += 1
    ds.save(['file_ingit', 'file_deleted'], to_git=True)
    # and add symlinks (untracked and in git)
    if check_symlink_capability(
        ds.pathobj / '_dummy', ds.pathobj / '_dummy_target'
    ):
        for i in ('symlinkuntracked', 'symlinkingit'):
            tpath = ds.pathobj / f'target_{i}'
            lpath = ds.pathobj / f'file_{i}'
            tpath.write_text(
                content_tmpl.format(i), encoding='utf-8')
            lpath.symlink_to(tpath)
            fcount += 1
    ds.save('file_symlinkingit', to_git=True)
    (ds.pathobj / 'file_deleted').unlink()
    return fcount, content_tmpl


def test_iter_gitworktree_basic_fp(existing_dataset, no_result_rendering):
    ds = existing_dataset
    fcount, content_tmpl = prep_fp_tester(ds)

    for ai in filter(
        lambda i: i.name.name.startswith('file_'),
        iter_gitworktree(ds.pathobj, fp=True)
    ):
        fcount -= 1
        if getattr(ai, 'fp', False):
            # for annexed files the fp can be an annex pointer file.
            # in the context of `iter_gitworktree` this is not a
            # recognized construct
            assert content_tmpl.format(
                ai.name.name[5:]) == ai.fp.read().decode() \
                or ai.name.name.startswith('file_annex')
        else:
            assert (ds.pathobj / ai.name).exists() is False
    assert not fcount


def test_iter_gitworktree_untracked_only(modified_dataset):
    p = modified_dataset.pathobj
    # only untracked files
    repo_items = list(iter_gitworktree(p, untracked='only'))
    assert all(f.name.name == 'file_u' for f in repo_items)
    # same report, but compressed to immediate directory children
    dir_items = list(iter_gitworktree(p, untracked='only', recursive='no'))
    assert set(f.name.parts[0] for f in repo_items) == \
        set(f.name.name for f in dir_items)
    # no wholly untracked directories in standard report
    assert not any(f.name.name == 'dir_u'
                   for f in iter_gitworktree(p, untracked='only'))
    # but this can be requested
    wholedir_items = list(iter_gitworktree(p, untracked='only-whole-dir'))
    assert any(f.name.name == 'dir_u' for f in wholedir_items)
    # smoke test remaining mode, test case doesn't cause difference
    assert any(f.name.name == 'dirempty_u' for f in wholedir_items)
    assert not any(f.name.name == 'dirempty_u'
                   for f in iter_gitworktree(p, untracked='only-no-empty-dir'))



def test_iter_gitworktree_pathspec(modified_dataset):
    p = modified_dataset.pathobj
    # query for any files that are set to go straight to Git. these are just
    # dotfiles in the default config
    items = list(iter_gitworktree(
        p,
        pathspecs=[':(attr:annex.largefiles=nothing)']))
    assert items
    assert all(str(i.name).startswith('.') for i in items)
    # glob-styles
    # first some that only give a top-level match
    assert len(list(iter_gitworktree(p, pathspecs=['file_a']))) == 1
    assert len(list(iter_gitworktree(p, pathspecs=[':(glob)*file_a']))) == 1
    # now some that match at any depth
    assert len(list(iter_gitworktree(p, pathspecs=['*file_a']))) == 2
    assert len(list(iter_gitworktree(p, pathspecs=[':(glob)**/file_a']))) == 2


def test_iter_submodules(modified_dataset):
    p = modified_dataset.pathobj
    all_sm = list(iter_submodules(p))
    assert all_sm
    assert all(sm.gittype == GitTreeItemType.submodule for sm in all_sm)
    assert all(str(sm.path.parent) == 'dir_sm' for sm in all_sm)
    assert sorted([str(sm.path.name) for sm in all_sm]) \
        == ['droppedsm_c', 'sm_c', 'sm_d', 'sm_m', 'sm_mu', 'sm_n',
            'sm_nm', 'sm_nmu', 'sm_u']
    # constrain by pathspec
    res = list(iter_submodules(p, pathspecs=['*/sm_c']))
    assert len(res) == 1
    assert res[0].name == PurePath('dir_sm', 'sm_c')
    # test negative condition
    res = list(iter_submodules(p, pathspecs=[':(exclude)*/sm_c']))
    assert len(res) == len(all_sm) - 1
    assert not any(r.name == PurePath('dir_sm', 'sm_c') for r in res)

    # test pathspecs matching inside submodules
    # baseline, pointing inside a submodule gives no matching results
    assert not list(iter_submodules(p, pathspecs=['dir_sm/sm_c/.datalad']))
    # we can discover the submodule that could have content that matches
    # the pathspec
    res = list(iter_submodules(p, pathspecs=['dir_sm/sm_c/.datalad'],
                               match_containing=True))
    assert len(res) == 1
    assert res[0].name == PurePath('dir_sm', 'sm_c')
    # if we use a wildcard that matches any submodule, we also get all of them
    # and this includes the dropped submodule, because iter_submodules()
    # make no assumptions on what this information will be used for
    res = list(iter_submodules(p, pathspecs=['*/.datalad'],
                               match_containing=True))
    assert len(res) == len(all_sm)


def test_iter_gitworktree_subm_recursion(modified_dataset):
    p = modified_dataset.pathobj
    nmu_items = list(iter_gitworktree(p / 'dir_sm' / 'sm_nmu',
                                      recursive='repository'))
    # doesn't matter how many exactly, but we expect multiple.
    # needed for the logic below
    assert len(nmu_items) > 1
    # and now from the top with recursion
    items = list(iter_gitworktree(p, recursive='submodules'))
    # we see all the submodule content
    assert all(
        any(i.name == PurePath('dir_sm', 'sm_nmu') / nmu.name for i in items)
        for nmu in nmu_items
    )
    # now we try listing only the 'nmu' submodule with a bunch of
    # equivalent pathspecs
    for ps in (
        # matches submodule directly
        ['dir_sm/sm_nmu'],
        # matches only inside the submodule
        # (test discovery of the submodule itself)
        ['dir_sm/sm_nmu/*'],
        [':(glob)dir_sm/sm_nmu/**'],
        [':(glob)dir_s?/*_nmu'],
    ):
        ps_items = iter_gitworktree(p, recursive='submodules', pathspecs=ps)
        # we see the submodule items, all of them, and only those
        assert [i.name for i in ps_items] == \
            [PurePath('dir_sm', 'sm_nmu') / i.name for i in nmu_items], \
            f'Mismatch for pathspec {ps!r}'



"""Data validation, coercion, and parameter documentation

This module provides a set of uniform classes to validate and document
particular aspects of inputs. In a nutshell, each of these
:class:`~datalad_next.constraints.Constraint` class:

- focuses on a specific aspect, such as data type coercion,
  or checking particular input properties
- is instantiated with a set of parameters to customize
  such an instance for a particular task
- performs its task by receiving an input via its ``__call__()``
  method
- provides default auto-documentation that can be customized
  by wrapping an instance in
  :class:`~datalad_next.constraints.WithDescription`

Individual ``Constraint`` instances can be combined with logical AND
(:class:`~datalad_next.constraints.AllOf`) and OR
(:class:`~datalad_next.constraints.AnyOf`) operations to form arbitrarily
complex constructs.

On (validation/coercion) error, instances raise
:class:`~datalad_next.constraints.ConstraintError`) via their
``raise_for()`` method. This approach to error reporting helps to communicate
standard (yet customizable) error messages, aids structured error reporting,
and is capable of communication the underlying causes of an error in full
detail without the need to generate long textual descriptions.

:class:`~datalad_next.constraints.EnsureCommandParameterization` is a
particular variant of a ``Constraint`` that is capable of validating a complete
parameterization of a command (or function), for each parameter individually,
and for arbitrary combinations of parameters. It puts a particular emphasis on
structured error reporting.


.. currentmodule:: datalad_next.constraints
.. autosummary::
   :toctree: generated

    Constraint
    AllOf
    AnyOf
    NoConstraint
    WithDescription

    ConstraintError
    CommandParametrizationError
    ParameterConstraintContext
    EnsureDataset
    DatasetParameter

    EnsureBool
    EnsureCallable
    EnsureChoice
    EnsureFloat
    EnsureHashAlgorithm
    EnsureDType
    EnsureInt
    EnsureKeyChoice
    EnsureNone
    EnsurePath
    EnsureStr
    EnsureStrPrefix
    EnsureRange
    EnsureValue

    EnsureIterableOf
    EnsureListOf
    EnsureTupleOf
    EnsureMapping
    EnsureGeneratorFromFileLike

    EnsureJSON
    EnsureURL
    EnsureParsedURL

    EnsureGitRefName
    EnsureRemoteName
    EnsureSiblingName

    EnsureCommandParameterization
"""
from .base import (
    AllOf,
    AnyOf,
    Constraint,
    DatasetParameter,
)
from .exceptions import (
    # this is the key type, almost all consuming code will want to
    # have this for `except` clauses
    ConstraintError,
    CommandParametrizationError,
    ParameterConstraintContext,
)
# expose constraints with direct applicability, but not
# base and helper classes
from .basic import (
    EnsureBool,
    EnsureCallable,
    EnsureChoice,
    EnsureFloat,
    EnsureHashAlgorithm,
    EnsureDType,
    EnsureInt,
    EnsureKeyChoice,
    EnsureNone,
    EnsurePath,
    EnsureStr,
    EnsureStrPrefix,
    EnsureRange,
    EnsureValue,
    NoConstraint,
)
from .compound import (
    EnsureIterableOf,
    EnsureListOf,
    EnsureTupleOf,
    EnsureMapping,
    EnsureGeneratorFromFileLike,
    WithDescription,
)
from .formats import (
    EnsureJSON,
    EnsureURL,
    EnsureParsedURL,
)

from .dataset import EnsureDataset

from .git import (
    EnsureGitRefName,
    EnsureRemoteName,
    EnsureSiblingName,
)

from .parameter import (
    EnsureCommandParameterization,
)



"""Special purpose exceptions

.. currentmodule:: datalad_next.exceptions
.. autosummary::
   :toctree: generated

   CapturedException
   IncompleteResultsError
   NoDatasetFound
"""
# we cannot have CommandError above, sphinx complains

# TODO rethink the purpose of this module and possibly
# make it about *external* custom exceptions
from datalad.runner.exception import CommandError
from datalad.support.exceptions import (
    CapturedException,
    IncompleteResultsError,
    NoDatasetFound,
)

# TODO REMOVE FOR V2.0 (they are specific to that module
from datalad_next.url_operations import (
    UrlOperationsRemoteError,
    UrlOperationsAuthenticationError,
    UrlOperationsAuthorizationError,
    UrlOperationsInteractionError,
    UrlOperationsResourceUnknown,
)



import pytest

from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest.mock import patch
from uuid import UUID

from datalad_next.commands import (
    ValidatedInterface,
    Parameter,
    eval_results,
)
from datalad_next.consts import on_windows
from .. import (
    ConstraintError,
    EnsureGeneratorFromFileLike,
    EnsureInt,
    EnsureJSON,
    EnsureListOf,
    EnsureMapping,
    EnsurePath,
    EnsureRange,
    EnsureStr,
    EnsureURL,
    EnsureValue,
)
from ..base import (
    AnyOf,
    Constraint,
)
from ..dataset import EnsureDataset
from ..parameter import EnsureCommandParameterization
from ..exceptions import ParameterConstraintContext


class EnsureAllUnique(Constraint):
    def __call__(self, value):
        if len(set(value)) < len(value):
            self.raise_for(value, 'not all values are unique')
        return value


class BasicCmdValidator(EnsureCommandParameterization):
    url_constraint = EnsureURL(required=['scheme'])
    url2path_constraint = EnsureMapping(
        key=url_constraint, value=EnsurePath(),
        delimiter='\t'
    )
    spec_item_constraint = url2path_constraint | url_constraint \
        | (EnsureJSON() & url2path_constraint)

    spec_constraint = AnyOf(
        EnsureListOf(spec_item_constraint),
        EnsureGeneratorFromFileLike(spec_item_constraint),
        spec_item_constraint,
    )

    def __init__(self, **kwargs):
        # this is the key bit: a mapping of parameter names to validators
        super().__init__(
            dict(
                spec=self.spec_constraint,
                p1=EnsureInt() | EnsureStr(),
            ),
            **kwargs
        )


class SophisticatedCmdValidator(BasicCmdValidator):
    def _check_unique_values(self, **kwargs):
        try:
            EnsureAllUnique()(kwargs.values())
        except ConstraintError as e:
            self.raise_for(
                kwargs,
                e.msg,
            )

    def _check_sum_range(self, p1, p2):
        try:
            EnsureRange(min=3)(p1 + p2)
        except ConstraintError:
            self.raise_for(
                dict(p1=p1, p2=p2),
                "it's too small"
            )

    def _limit_sum_range(self, p1, p2):
        # random example of a joint constraint that modifies the parameter
        # set it is given
        return dict(p1=p1, p2=min(p2, 100 - p1 - p2))

    def __init__(self):
        # this is the key bit: a mapping of parameter names to validators
        super().__init__(
            # implementation example of a higher-order constraint
            joint_constraints={
                ParameterConstraintContext(('p1', 'p2'), 'identity'):
                    self._check_unique_values,
                ParameterConstraintContext(('p1', 'p2'), 'sum'):
                    self._check_sum_range,
                ParameterConstraintContext(('p1', 'p2'), 'sum-limit'):
                    self._limit_sum_range,
            },
        )


class BrokenJointValidation(SophisticatedCmdValidator):
    def joint_validation(self, params, on_error):
        res = super().joint_validation(params, on_error)
        # remove any report, and it should trigger a RuntimeError on return
        res.popitem()
        return res


class CmdWithValidation(ValidatedInterface):
    # this is of little relevance, no validation configured here
    _params_ = dict(spec=Parameter(args=('spec',), nargs='+'))

    _validator_ = BasicCmdValidator()

    # command implementation that only validates and returns the outcome
    @staticmethod
    @eval_results
    def __call__(spec, p1='one', p2='two'):
        yield dict(
            action='cmd_with_validation',
            # list() consumes any potential generator
            spec=list(spec),
            status='ok',
        )


def test_multi_validation():
    val = BasicCmdValidator()
    # we break the parameter specification, and get a ConstraintError
    with pytest.raises(ConstraintError) as e:
        val(dict(spec='5'))
    # but actually, it is a ConstraintErrors instance, and we get the
    # violation exceptions within the context in which they occurred.
    # here this is a parameter name
    errors = e.value.errors
    assert len(errors) == 1
    ctx = ParameterConstraintContext(('spec',))
    assert ctx in errors
    assert errors[ctx].constraint == BasicCmdValidator.spec_constraint
    assert errors[ctx].value == '5'

    # now we trigger a higher-order error, and receive multiple reports
    val = SophisticatedCmdValidator()
    # but first a quick check if it behaves will with valid input
    valid_input = dict(spec='http://example.com', p1=1, p2=2)
    assert val(valid_input) == valid_input
    with pytest.raises(ConstraintError) as e:
        val(dict(spec='5', p1=1, p2=1), on_error='raise-at-end')
    errors = e.value.errors
    assert len(errors) == 3
    # the spec-param-only error
    assert errors.messages[0].startswith('does not match any of')
    # higher-order issue traces (their order is deterministic)
    assert 'not all values are unique' == errors.messages[1]
    assert 'p1, p2 (identity)' == errors.context_labels[1]
    assert 'p1, p2 (sum)' in errors.context_labels
    # and now against, but with stop-on-first-error
    with pytest.raises(ConstraintError) as e:
        val(dict(spec='5', p1=1, p2=1), on_error='raise-early')
    errors = e.value.errors
    # and we only get one!
    assert len(errors) == 1
    # the spec-param-only error
    assert errors.messages[0].startswith('does not match any of')
    assert 'not all values are unique' not in errors.messages
    # now we do it again, but with a valid spec, such that the first
    # and only error is a higher order error
    with pytest.raises(ConstraintError) as e:
        val(dict(spec=5, p1=1, p2=1), on_error='raise-early')
    errors = e.value.errors
    assert len(errors) == 1
    assert 'not all values are unique' == errors.messages[0]
    assert 'p1, p2 (identity)' == errors.context_labels[0]

    # a single-parameter validation error does not lead to a crash
    # in higher-order validation, instead the latter is performed
    # when a require argument could not be provided
    with pytest.raises(ConstraintError) as e:
        # p1 must be int|str
        val(dict(spec=5, p1=None, p2=1), on_error='raise-at-end')


def test_invalid_multi_validation():
    val = BrokenJointValidation()
    # this works for the underlying validator, but BrokenJointValidation
    # butchers the result, which must be detected
    valid_input = dict(spec='http://example.com', p1=1, p2=2)
    with pytest.raises(RuntimeError):
        val(valid_input)


def test_cmd_with_validation(no_result_rendering):
    target_urls = ['http://example.com', 'file:///dev/null']
    target_url_path_maps = [
        {'http://example.com': Path('some/dir/file')},
        {'file:///dev/null': Path('/dev/null')},
    ]
    json_lines = '{"http://example.com":"some/dir/file"}\n' \
                 '{"file:///dev/null":"/dev/null"}'

    for input, target in (
        # perfect input
        (target_urls, target_urls),
        (target_url_path_maps, target_url_path_maps),
        # actual invput conversion
        ([{'http://example.com': 'some/dir/file'},
          {'file:///dev/null': '/dev/null'}],
         target_url_path_maps),
        # custom mapping syntax
        (['http://example.com\tsome/dir/file',
          'file:///dev/null\t/dev/null'],
         target_url_path_maps),
        # JSON lines
        (['{"http://example.com":"some/dir/file"}',
          '{"file:///dev/null":"/dev/null"}'],
         target_url_path_maps),
        # from file with JSON lines
        (StringIO(json_lines), target_url_path_maps),
    ):
        res = CmdWithValidation.__call__(
            input,
            return_type='item-or-list',
        )
        assert 'spec' in res
        assert res['spec'] == target

    # read from file
    if not on_windows:
        # on windows the write-rewind-test logic is not possible
        # (PermissionError) -- too lazy to implement a workaround
        with NamedTemporaryFile('w+') as f:
            f.write(json_lines)
            f.seek(0)
            res = CmdWithValidation.__call__(
                f.name,
                return_type='item-or-list',
            )
            assert res['spec'] == target_url_path_maps

    with patch("sys.stdin", StringIO(json_lines)):
        res = CmdWithValidation.__call__(
            '-',
            return_type='item-or-list',
        )
        assert res['spec'] == target_url_path_maps

    # and now something that fails
    # TODO error reporting should be standardized (likely) on an explicit
    # and dedicated exception type
    # https://github.com/datalad/datalad/issues/7167
    with pytest.raises(ValueError):
        CmdWithValidation.__call__(
            'unsupported',
            return_type='item-or-list',
        )

    # no call with a required argument missing
    with pytest.raises(ValueError) as e:
        CmdWithValidation.__call__()
    exc_rendering = str(e.value)
    # must label the issue correctly
    assert 'missing required argument' in exc_rendering
    # must identify the missing argument
    assert 'spec=<no value>' in exc_rendering


#
# test dataset tailoring
#

class EnsureUUID(Constraint):
    def __call__(self, value):
        return UUID(value)


class EnsureDatasetID(EnsureUUID):
    """Makes sure that something is a dataset ID (UUID), or the dataset UUID
    of a particular dataset when tailored"""
    def for_dataset(self, dsarg):
        return EnsureValue(UUID(dsarg.ds.id))


class DsTailoringValidator(EnsureCommandParameterization):
    def __init__(self, **kwargs):
        # this is the key bit: a mapping of parameter names to validators
        super().__init__(
            dict(
                dataset=EnsureDataset(),
                id=EnsureDatasetID(),
            ),
            **kwargs
        )


def test_constraint_dataset_tailoring(existing_dataset):
    proper_uuid = '152f4fc0-b444-11ed-a9cb-701ab88b716c'
    no_uuid =     '152f4fc0------11ed-a9cb-701ab88b716c'
    # no tailoring works as expected
    val = DsTailoringValidator()
    assert val(dict(id=proper_uuid)) == dict(id=UUID(proper_uuid))
    with pytest.raises(ValueError):
        val(dict(id=no_uuid))

    # adding a dataset to the mix does not change a thing re the uuid
    ds = existing_dataset
    res = val(dict(dataset=ds.pathobj, id=proper_uuid))
    assert res['id'] == UUID(proper_uuid)
    assert res['dataset'].ds == ds

    # and we can still break it
    with pytest.raises(ValueError):
        val(dict(dataset=ds.pathobj, id=no_uuid))

    # now with tailoring the UUID checking to a particular dataset.
    # it is enabled via parameter, because it is a use case specific
    # choice, not a mandate
    target_uuid = ds.id
    tailoring_val = DsTailoringValidator(
        tailor_for_dataset=dict(id='dataset'),
    )
    # no uuid is still an issue
    with pytest.raises(ValueError):
        tailoring_val(dict(dataset=ds.pathobj, id=no_uuid))
    # what was good enough above (any UUID), no longer is
    with pytest.raises(ValueError):
        tailoring_val(dict(dataset=ds.pathobj, id=proper_uuid))

    # only the actual dataset's UUID makes it past the gates
    res = val(dict(dataset=ds.pathobj, id=target_uuid))
    assert res['id'] == UUID(target_uuid)
    assert res['dataset'].ds == ds
    # the order in the kwargs does not matter
    assert val(dict(id=target_uuid, dataset=ds.pathobj))['id'] == \
        UUID(target_uuid)

    # but when no dataset is being provided (and the dataset-constraint
    # allows for that), no tailoring is performed
    assert tailoring_val(dict(id=proper_uuid))['id'] == UUID(proper_uuid)
    # but still no luck with invalid args
    with pytest.raises(ValueError):
        val(dict(dataset=ds.pathobj, id=no_uuid))



"""``dl+archive:`` archive member locator"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import PurePosixPath
import re

from .annexkey import AnnexKey
from .enums import ArchiveType


# be relatively permissive
_recognized_urls = re.compile(r'^dl\+archive:(?P<key>.*)#(?P<props>.*)')
# each archive member is identified by a (relative) path inside
# the archive.
_archive_member_props = re.compile(
    # a path may contain any char but '&'
    # TODO check that something in the machinery ensures proper
    # quoting
    'path=(?P<path>[^&]+)'
    # size info (in bytes) is optional
    '(&size=(?P<size>[0-9]+)|)'
    # archive type label is optional
    '(&atype=(?P<atype>[a-z0-9]+)|)'
)


@dataclass
class ArchivistLocator:
    """Representation of a ``dl+archive:`` archive member locator

    These locators are used by the ``datalad-archives`` and ``archivist``
    git-annex special remotes. They identify a member of a archive that is
    itself identified by an annex key.

    Each member is annotated with its size (in bytes). Optionally,
    the file format type of the archive can be annotated too.

    Syntax of ``dl+archives:`` locators
    -----------------------------------

    The locators the following minimal form::

        dl+archive:<archive-key>#path=<path-in-archive>

    where ``<archive-key>`` is a regular git-annex key of an archive file,
    and ``<path-in-archive>`` is a POSIX-style relative path pointing to
    a member within the archive.

    Two optional, additional attributes ``size`` and ``atype`` are recognized
    (only ``size`` is also understood by the ``datalad-archives``
    special remote).

    ``size`` declares the size of the (extracted) archive member in bytes::

        dl+archive:<archive-key>#path=<path-in-archive>&size=<size-in-bytes>

    ``atype`` declares the type of the containing archive using a label.
    Currently recognized labels are ``tar`` (a TAR archive, compressed or not),
    and ``zip`` (a ZIP archive). See
    :class:`~datalad_next.types.enums.ArchiveType` for all recognized labels.

    If no type information is given, :func:`ArchivistLocator.from_str()` will
    try to determine the archive type from the archive key (via ``*E``-type
    git-annex backends, such as DataLad's default ``MD5E``).

    The order in the fragment part of the URL (after ``#``) is significant.
    ``path`` must come first, followed by ``size`` or ``atype``. If both
    ``size`` and ``atype`` are present, ``size`` must be declared first. A
    complete example of a URL is::

        dl+archive:MD5-s389--e9f624eb778e6f945771c543b6e9c7b2#path=dir/file.csv&size=234&atype=tar
    """
    akey: AnnexKey
    member: PurePosixPath
    size: int | None = None
    # datalad-archives did not have the type info, we want to be
    # able to handle those too, make optional
    atype: ArchiveType | None = None

    def __str__(self) -> str:
        return 'dl+archive:{akey}#path={member}&size={size}{atype}'.format(
            akey=self.akey,
            # TODO needs quoting?
            member=self.member,
            size=self.size,
            atype=f'&atype={self.atype.value}' if self.atype else '',
        )

    @classmethod
    def from_str(cls, url: str):
        """Return ``ArchivistLocator`` from ``str`` form"""
        url_match = _recognized_urls.match(url)
        if not url_match:
            raise ValueError('Unrecognized dl+archives locator syntax')
        url_matched = url_match.groupdict()
        # convert to desired type
        akey = AnnexKey.from_str(url_matched['key'])

        # archive member properties
        props_match = _archive_member_props.match(url_matched['props'])
        if not props_match:
            # without at least a 'path' there is nothing we can do here
            raise ValueError(
                'dl+archives locator contains invalid archive member '
                f'specification: {url_matched["props"]!r}')
        props_matched = props_match.groupdict()
        amember_path = PurePosixPath(props_matched['path'])
        if amember_path.is_absolute():
            raise ValueError(
                'dl+archives locator contains absolute archive member path')
        if '..' in amember_path.parts:
            raise ValueError(
                'dl+archives locator archive member path contains ".."')

        # size is optional, regex ensure that it is an int
        size = props_matched.get('size')
        if size is not None:
            size = int(size)

        # archive type, could be None
        atype = props_matched.get('atype')
        if atype is not None:
            # if given, most be known type
            try:
                atype = getattr(ArchiveType, atype)
            except AttributeError as e:
                raise ValueError(
                    'dl+archives locator archive type unrecognized') from e

        if atype is None and akey.backend.endswith('E'):
            # try by key name extension
            suf = PurePosixPath(akey.name).suffixes
            if '.zip' == suf[-1]:
                atype = ArchiveType.zip
            elif '.tar' in suf:
                atype = ArchiveType.tar
            elif '.tgz' in suf:
                atype = ArchiveType.tar

        return cls(
            akey=akey,
            member=amember_path,
            size=size,
            atype=atype,
        )



from __future__ import annotations

import gzip
import pytest
import requests

from datalad_next.tests import skipif_no_network

from ..any import AnyUrlOperations
from ..http import (
    HttpUrlOperations,
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)


def test_http_url_operations(credman, httpbin, tmp_path):
    hbsurl = httpbin['standard']
    hbscred = (
        'hbscred',
        dict(user='mike', secret='dummy', type='user_password',
             realm=f'{hbsurl}/Fake Realm'),
    )

    credman.set(hbscred[0], **hbscred[1])
    ops = HttpUrlOperations()
    # authentication after redirect
    target_url = f'{hbsurl}/basic-auth/mike/dummy'
    props = ops.stat(f'{hbsurl}/redirect-to?url={target_url}')
    # we get the resolved URL after redirect back
    assert props['url'] == target_url
    # same again, but credentials are wrong
    target_url = f'{hbsurl}/basic-auth/mike/WRONG'
    with pytest.raises(UrlOperationsRemoteError):
        ops.stat(f'{hbsurl}/redirect-to?url={target_url}')
    # make sure we get the size info
    assert ops.stat(f'{hbsurl}/bytes/63')['content-length'] == 63

    # download
    # SFRUUEJJTiBpcyBhd2Vzb21l == 'HTTPBIN is awesome'
    props = ops.download(f'{hbsurl}/base64/SFRUUEJJTiBpcyBhd2Vzb21l',
                         tmp_path / 'mydownload',
                         hash=['md5'])
    assert (tmp_path / 'mydownload').read_text() == 'HTTPBIN is awesome'

    # 404s
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.stat(f'{hbsurl}/status/404')
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.download(f'{hbsurl}/status/404', tmp_path / 'dontmatter')


def test_custom_http_headers_via_config(datalad_cfg):
    for k, v in (
            ('datalad.url-handler.http.*.class',
             'datalad_next.url_operations.http.HttpUrlOperations'),
            ('datalad.url-handler.http.*.kwargs',
             '{"headers": {"X-Funky": "Stuff"}}'),
    ):
        datalad_cfg.set(k, v, scope='global', reload=False)
    datalad_cfg.reload()
    auo = AnyUrlOperations()
    huo = auo._get_handler(f'http://example.com')
    assert huo._headers['X-Funky'] == 'Stuff'


@skipif_no_network
def test_transparent_decompression(tmp_path):
    # this file is offered with transparent compression/decompression
    # by the github webserver
    url = 'https://raw.githubusercontent.com/datalad/datalad-next/' \
          'd0c4746425a48ef20e3b1c218e68954db9412bee/pyproject.toml'
    dpath = tmp_path / 'test.txt'
    ops = HttpUrlOperations()
    ops.download(from_url=url, to_path=dpath)

    # make sure it ends up on disk uncompressed
    assert dpath.read_text() == \
        '[build-system]\nrequires = ["setuptools >= 43.0.0", "wheel"]\n'


@skipif_no_network
def test_compressed_file_stay_compressed(tmp_path):
    # this file is offered with transparent compression/decompression
    # by the github webserver, but is also actually gzip'ed
    url = \
        'https://github.com/datalad/datalad-neuroimaging/raw/' \
        '05b45c8c15d24b6b894eb59544daa17159a88945/' \
        'datalad_neuroimaging/tests/data/files/nifti1.nii.gz'

    # first confirm validity of the test approach, opening an
    # uncompressed file should raise an exception
    with pytest.raises(gzip.BadGzipFile):
        testpath = tmp_path / 'uncompressed'
        testpath.write_text('some')
        with gzip.open(testpath, 'rb') as f:
            f.read(1000)

    # and now with a compressed file
    dpath = tmp_path / 'test.nii.gz'
    ops = HttpUrlOperations()
    ops.download(from_url=url, to_path=dpath)
    # make sure it ends up on disk compressed!
    with gzip.open(dpath, 'rb') as f:
        f.read(1000)


def test_size_less_progress_reporting(http_server, monkeypatch):
    test_file = (http_server.path / 'test.bin').open('wb')
    test_file.seek(100000)
    test_file.write(b'a')
    test_file.close()

    r = requests.get(http_server.url + '/test.bin', stream=True)
    del r.headers['content-length']

    logs = []
    # patch the log_progress() used in http.py
    def catch_progress(*_, **kwargs):
        logs.append(kwargs)

    import datalad_next.url_operations.base
    monkeypatch.setattr(datalad_next.url_operations.base, 'log_progress', catch_progress)

    http_handler = HttpUrlOperations()
    http_handler._stream_download_from_request(r, None)
    assert any('update' in kwargs for kwargs in logs)
    assert any(('total', None) in kwargs.items() for kwargs in logs)


def test_header_adding():
    default_headers = dict(key_1='value_1')
    added_headers = dict(key_2='value_2')
    url_ops = HttpUrlOperations(headers=default_headers)
    assert 'key_1' in url_ops.get_headers()

    # ensure that header entries from `headers` show up in result
    combined_keys = {'key_1', 'key_2'}
    result_1 = url_ops.get_headers(headers=dict(added_headers))
    assert combined_keys.issubset(set(result_1))

    # ensure that `headers` did not change the stored headers
    result_2 = url_ops.get_headers()
    assert 'key_2' not in set(result_2)


def test_delete_method(httpbin):
    url_base = httpbin['standard']
    url_ops = HttpUrlOperations()
    url_ops.delete(f'{url_base}/delete', timeout=1.0)


def test_delete_timeout(httpbin):
    url_base = httpbin['standard']
    url_ops = HttpUrlOperations()
    with pytest.raises(TimeoutError):
        url_ops.delete(f'{url_base}/delay/2', timeout=1.0)



"""Add support for export to WebDAV remotes to ``push()``

This approach generally works for any special remote configured
with ``exporttree=yes``, but is only tested for ``type=webdav``.
A smooth operation requires automatic deployment of
credentials. Support for that is provide and limited by
the capabilities of ``needs_specialremote_credential_envpatch()``.
"""

import logging
from typing import (
    Dict,
    Generator,
    Iterable,
    Optional,
    Union,
)

import datalad.core.distributed.push as mod_push
from datalad_next.constraints import EnsureChoice
from datalad_next.exceptions import CapturedException
from datalad_next.commands import Parameter
from datalad_next.datasets import (
    LegacyAnnexRepo as AnnexRepo,
    Dataset,
)
from datalad_next.utils import (
    CredentialManager,
    get_specialremote_credential_envpatch,
    get_specialremote_credential_properties,
    needs_specialremote_credential_envpatch,
    patched_env,
)
from . import apply_patch


lgr = logging.getLogger('datalad.core.distributed.push')


def _is_export_remote(remote_info: Optional[Dict]) -> bool:
    """Check if remote_info is valid and has exporttree set to "yes"

    Parameters
    ----------
    remote_info: Optional[Dict]
        Optional dictionary the contains git annex special.

    Returns
    -------
    bool
        True if exporttree key is contained in remote_info and is set to yes,
        else False.
    """
    if remote_info is not None:
        return remote_info.get("exporttree") == "yes"
    return False


def _get_credentials(ds: Dataset,
                     remote_info: Dict
                     ) -> Optional[Dict]:

    # Check for credentials
    params = {
        "type": remote_info.get("type"),
        "url": remote_info.get("url")
    }
    credentials = None
    credential_properties = get_specialremote_credential_properties(params)
    if credential_properties:
        # TODO: lower prio: factor this if clause out, also used in
        #  create_sibling_webdav.py
        credential_manager = CredentialManager(ds.config)
        credentials = (credential_manager.query(
            _sortby='last-used',
            **credential_properties) or [(None, None)])[0][1]
    return credentials


def get_export_records(repo: AnnexRepo) -> Generator:
    """Read exports that git-annex recorded in its 'export.log'-file

    Interpret the lines in export.log. Each line has the following structure:

        time-stamp " " source-annex-uuid ":" destination-annex-uuid " " treeish

    Parameters
    ----------
    repo: AnnexRepo
        The annex repo from which exports should be determined

    Returns
    -------
    Generator
        Generator yielding one dictionary for each export entry in git-annex.
        Each dictionary contains the keys: "timestamp", "source-annex-uuid",
        "destination-annex-uuid", "treeish". The timestamp-value is a float,
        all other values are strings.
    """
    try:
        # XXX when this is changed to `call_git()`, make sure to use
        # `force_c_locale=True`
        for line in repo.call_git_items_(["cat-file", "blob", "git-annex:export.log"]):
            result_dict = dict(zip(
                [
                    "timestamp",
                    "source-annex-uuid",
                    "destination-annex-uuid",
                    "treeish"
                ],
                line.replace(":", " ").split()
            ))
            result_dict["timestamp"] = float(result_dict["timestamp"][:-1])
            yield result_dict
    except mod_push.CommandError as command_error:
        # Some errors indicate that there was no export yet.
        # May depend on Git version
        expected_errors = (
            "fatal: Not a valid object name git-annex:export.log",
            "fatal: path 'export.log' does not exist in 'git-annex'", # v2.36
        )
        if command_error.stderr.strip() in expected_errors:
            return
        raise


def _get_export_log_entry(repo: AnnexRepo,
                          target_uuid: str
                          ) -> Optional[Dict]:
    target_entries = [
        entry
        for entry in repo.get_export_records()
        if entry["destination-annex-uuid"] == target_uuid]

    if not target_entries:
        return None
    return sorted(target_entries, key=lambda e: e["timestamp"])[-1]


def _is_valid_treeish(repo: AnnexRepo,
                      export_entry: Dict,
                      ) -> bool:

    # Due to issue https://github.com/datalad/datalad-next/issues/39
    # fast-forward validation has to be re-designed.
    return True
    #for line in repo.call_git_items_(["log", "--pretty=%H %T"]):
    #    commit_hash, treeish = line.split()
    #    if treeish == export_entry["treeish"]:
    #        return True
    #return False


def _transfer_data(repo: AnnexRepo,
                   ds: Dataset,
                   target: str,
                   content: Iterable,
                   data: str,
                   force: Optional[str],
                   jobs: Optional[Union[str, int]],
                   res_kwargs: Dict,
                   got_path_arg: bool
                   ) -> Generator:

    target_uuid, remote_info = ([
        (uuid, info) for uuid, info in repo.get_special_remotes().items()
        if info.get("name") == target] or [(None, None)])[0]

    if not _is_export_remote(remote_info):
        yield from mod_push._push_data(
            ds,
            target,
            content,
            data,
            force,
            jobs,
            res_kwargs.copy(),
            got_path_arg=got_path_arg,
        )
        return

    from datalad.interface.results import annexjson2result

    # TODO:
    #  - check for configuration entries, e.g. what to export

    lgr.debug(f"Exporting HEAD of {ds} to remote {remote_info}")

    if ds.config.getbool('remote.{}'.format(target), 'annex-ignore', False):
        lgr.debug(
            "Target '%s' is set to annex-ignore, exclude from data-export.",
            target)
        return

    if force not in ("all", "export"):
        export_entry = _get_export_log_entry(repo, target_uuid)
        if export_entry:
            if export_entry["source-annex-uuid"] != repo.uuid:
                yield dict(
                    **res_kwargs,
                    status="error",
                    message=f"refuse to export to {target}, because the "
                            f"last known export came from another repo "
                            f"({export_entry['source-annex-uuid']}). Use "
                            f"--force=export to enforce the export anyway.")
                return
            if not _is_valid_treeish(repo, export_entry):
                yield dict(
                    **res_kwargs,
                    status="error",
                    message=f"refuse to export to {target}, because the "
                            f"current state is not a fast-forward of the "
                            f"last known exported state. Use "
                            f"--force=export to enforce the export anyway.")
                return

    credentials = _get_credentials(ds, remote_info)

    # If we have credentials, check whether we require an environment patch
    env_patch = {}
    remote_type = remote_info.get("type")
    if credentials and needs_specialremote_credential_envpatch(remote_type):
        env_patch = get_specialremote_credential_envpatch(
            remote_type,
            credentials)

    res_kwargs['target'] = target

    with patched_env(**env_patch):
        try:
            for result in repo._call_annex_records_items_(
                [
                    "export", "HEAD",
                    "--to", target
                ],
                progress=True
            ):
                result_adjusted = \
                    annexjson2result(result, ds, **res_kwargs)
                # annexjson2result overwrites 'action' with annex' 'command',
                # even if we provided our 'action' within res_kwargs. Therefore,
                # change afterwards instead:
                result_adjusted['action'] = "copy"
                yield result_adjusted

        except mod_push.CommandError as cmd_error:
            ce = CapturedException(cmd_error)
            yield {
                **res_kwargs,
                "action": "copy",
                "status": "error",
                "message": str(ce),
                "exception": ce
            }


apply_patch('datalad.core.distributed.push', None, '_transfer_data',
            _transfer_data)
lgr.debug(
    "Patching datalad.core.distributed.push.Push docstring and parameters")
mod_push.Push.__doc__ += """\


    The following feature is added by the datalad-next extension:

    If a target is a git-annex special remote that has "exporttree" set to
    "yes", push will call 'git-annex export' to export the current HEAD to the
    remote target. This will usually result in a copy of the file tree, to which
    HEAD refers, on the remote target. A git-annex special remote with
    "exporttree" set to "yes" can, for example, be created with the datalad
    command "create-sibling-webdav" with the option "--mode=filetree" or 
    "--mode=filetree-only".
"""
mod_push.Push._params_["force"] = Parameter(
    args=("-f", "--force",),
    doc="""force particular operations, possibly overruling safety
    protections or optimizations: use --force with git-push ('gitpush');
    do not use --fast with git-annex copy ('checkdatapresent'); force an
    annex export (to git annex remotes with "exporttree" set to "yes");
    combine all force modes ('all').""",
    constraints=EnsureChoice(
        'all', 'gitpush', 'checkdatapresent', 'export', None))


from datalad.interface.base import build_doc
mod_push.Push.__call__.__doc__ = None
mod_push.Push = build_doc(mod_push.Push)

apply_patch(
    'datalad_next.datasets', 'LegacyAnnexRepo', 'get_export_records',
    get_export_records,
    msg="Patching datalad.support.AnnexRepo.get_export_records (new method)",
    expect_attr_present=False,
)



from io import StringIO
import pytest

from datalad_next.commands import Parameter
from datalad_next.utils import chpwd

from ..base import DatasetParameter
from ..basic import (
    EnsureInt,
    EnsureStr,
    NoConstraint,
)
from ..compound import EnsureGeneratorFromFileLike
from ..dataset import EnsureDataset
from ..exceptions import (
    ConstraintError,
    NoDatasetFound,
)
from ..formats import (
    EnsureJSON,
    EnsureURL,
    EnsureParsedURL,
)
from ..git import (
    EnsureGitRefName,
    EnsureRemoteName
)
from ..parameter_legacy import EnsureParameterConstraint


def test_EnsureGitRefName():
    assert EnsureGitRefName().short_description() == '(single-level) Git refname'
    # standard branch name must work
    assert EnsureGitRefName()('main') == 'main'
    # normalize is on by default
    assert EnsureGitRefName()('/main') == 'main'
    with pytest.raises(ValueError):
        EnsureGitRefName(normalize=False)('/main')
    assert EnsureGitRefName(normalize=False)('main') == 'main'
    # no empty
    with pytest.raises(ValueError):
        EnsureGitRefName()('')
    with pytest.raises(ValueError):
        EnsureGitRefName()(None)
    # be able to turn off onelevel
    with pytest.raises(ValueError):
        EnsureGitRefName(allow_onelevel=False)('main')
    assert EnsureGitRefName(allow_onelevel=False)(
        'refs/heads/main') == 'refs/heads/main'
    # refspec pattern off by default
    with pytest.raises(ValueError):
        EnsureGitRefName()('refs/heads/*')
    assert EnsureGitRefName(refspec_pattern=True)(
        'refs/heads/*') == 'refs/heads/*'


def test_EnsureRemoteName(existing_dataset):
    # empty sibling name must raise
    with pytest.raises(ValueError):
        EnsureRemoteName()('')
    assert EnsureRemoteName().short_description() == 'Name of a remote'
    assert EnsureRemoteName(
        known=True).short_description() == 'Name of a known remote'
    assert EnsureRemoteName(
        known=False).short_description() == 'Name of a not-yet-known remote'
    ds = existing_dataset
    c = EnsureRemoteName(known=False)
    tc = c.for_dataset(DatasetParameter(None, ds))
    assert tc('newremotename') == 'newremotename'
    # add a remote
    ds._repo.add_remote('my-remote', 'here')
    # check should fail when it shouldn't exist
    with pytest.raises(ValueError):
        tc('my-remote')
    # should work when it should exist
    c = EnsureRemoteName(known=True)
    tc = c.for_dataset(DatasetParameter(None, ds))
    assert tc('my-remote') == 'my-remote'
    # but fail with non-existing remote
    with pytest.raises(ValueError) as e:
        tc('not-my-remote')
    assert str(e.value) == "is not a known remote"
    # return sibling name with no existence checks
    assert EnsureRemoteName()('anything') == 'anything'


def test_EnsureParameterConstraint():
    # most basic case, no value constraint
    c = EnsureParameterConstraint(NoConstraint())
    # invalid name
    with pytest.raises(ValueError):
        c({'4way': 123})
    assert c('so1230_s82me=value') == dict(so1230_s82me='value')
    # now some from a standard Parameter declaration
    c = EnsureParameterConstraint.from_parameter(
        Parameter(), 'whateverdefault')
    assert c('some=value') == dict(some='value')

    # want a bool
    c = EnsureParameterConstraint.from_parameter(
        Parameter(action="store_true"),
        False)
    assert c('some=off') == dict(some=False)
    with pytest.raises(ValueError):
        c('some=5')
    c = EnsureParameterConstraint.from_parameter(
        # argparse specific choice declaration without
        # any constraint
        Parameter(choices=['a', 'b']),
        # but specifically use a default that is not a valid choice
        None)
    assert c('choice=a') == dict(choice='a')
    # default is valid too
    assert c({'choice': None}) == dict(choice=None)
    # multi-item values
    c = EnsureParameterConstraint.from_parameter(
        Parameter(nargs=2),
        (None, None))
    assert c({'some': [3, 4]}) == dict(some=[3, 4])
    with pytest.raises(ValueError):
        c({'some': 3})
    with pytest.raises(ValueError):
        c({'some': [3, 4, 5]})
    # one or more items
    c = EnsureParameterConstraint.from_parameter(
        Parameter(nargs='*'),
        None)
    # always prefers a list, no item type conversion by default
    assert c('some=5') == dict(some=['5'])
    assert c({'some': [5, 2]}) == dict(some=[5, 2])
    # empty ok
    assert c({'some': []}) == dict(some=[])
    # at least one item
    c = EnsureParameterConstraint.from_parameter(
        Parameter(nargs='+', constraints=EnsureInt()),
        None)
    assert c('some=5') == dict(some=[5])
    assert c({'some': [5, 2]}) == dict(some=[5, 2])
    # empty not ok
    with pytest.raises(ValueError):
        c({'some': []})
    # complex case of iterables of length 2
    c = EnsureParameterConstraint.from_parameter(
        Parameter(nargs=2, constraints=EnsureInt(), action='append'),
        None)
    # no iterable does not violate
    assert c({'some': []}) == dict(some=[])
    assert c({'some': [[3, 2]]}) == dict(some=[[3, 2]])
    assert c({'some': [[3, 2], [5, 4]]}) == dict(some=[[3, 2], [5, 4]])
    # length mismatch
    with pytest.raises(ValueError):
        c({'some': [[3, 2], [1]]})
    # no iterable
    with pytest.raises(ValueError):
        c({'some': [3, [1, 2]]})
    with pytest.raises(ValueError):
        c({'some': 3})
    # overwrite an item constraint and nargs
    c = EnsureParameterConstraint.from_parameter(
        Parameter(nargs=2, constraints=EnsureInt(), action='append'),
        None,
        item_constraint=EnsureStr(),
        nargs=1)
    assert c({'some': ['5']}) == dict(some=['5'])
    # literal constraint label
    # this is no longer supported, but still works: test until removed
    with pytest.deprecated_call():
        c = EnsureParameterConstraint.from_parameter(
            Parameter(), 2, item_constraint='float')
    assert c('some=3') == dict(some=3.0)
    with pytest.raises(ValueError), \
            pytest.deprecated_call():
        EnsureParameterConstraint.from_parameter(
            Parameter(), 2, item_constraint='unknown')


def test_EnsureParameterConstraint_passthrough():
    c = EnsureParameterConstraint(EnsureInt(), passthrough=None)
    # rejects wrong ones
    with pytest.raises(ValueError):
        c('p=mike')
    # accepts correct ones
    assert c('p=5') == {'p': 5}
    # and passes through
    assert c(dict(p=None)) == {'p': None}
    # even when the actual value constraint would not
    with pytest.raises(ConstraintError):
        c.parameter_constraint.constraint(None)
    # setting is retrievable
    assert c.passthrough_value is None

    # now the "same" via from_parameter()
    c = EnsureParameterConstraint.from_parameter(
        Parameter(constraints=EnsureInt()),
        default=None)
    assert c(dict(p=None)) == {'p': None}
    assert c('p=5') == {'p': 5}


nested_json = """\
{"name": "Alexa", "wins": [["two pair", "4♠"], ["two pair", "9♠"]]}
"""
nested_json_decoded = {
    "name": "Alexa",
    "wins": [["two pair", "4♠"],
             ["two pair", "9♠"]],
}
invalid_json = """\
{"name": BOOM!}
"""


def test_EnsureJSONLines():
    constraint = EnsureGeneratorFromFileLike(EnsureJSON())

    assert 'items of type "JSON" read from a file-like' \
        ==  constraint.short_description()

    # typical is "object", but any valid JSON value type must work
    assert list(constraint(StringIO("5"))) == [5]
    # unicode must work
    uc = "ΔЙקم๗あ"
    assert list(constraint(StringIO(f'"{uc}"'))) == [uc]
    assert list(constraint(StringIO(nested_json))) == [nested_json_decoded]

    with pytest.raises(ValueError) as e:
        list(constraint(StringIO(f'{nested_json}\n{invalid_json}')))


url_testcases = {
    "http://www.google.com": ['netloc','scheme',],
    "https://www.google.com": ['netloc','scheme',],
    "http://google.com": ['netloc','scheme',],
    "https://google.com": ['netloc','scheme',],
    "www.google.com": ['path',],
    "google.com": ['path',],
    "http://www.google.com/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "https://www.google.com/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "http://google.com/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "https://google.com/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "www.google.com/~as_db3.2123/134-1a": ['path',],
    "google.com/~as_db3.2123/134-1a": ['path',],
    # .co.uk top level
    "http://www.google.co.uk": ['netloc','scheme',],
    "https://www.google.co.uk": ['netloc','scheme',],
    "http://google.co.uk": ['netloc','scheme',],
    "https://google.co.uk": ['netloc','scheme',],
    "www.google.co.uk": ['path',],
    "google.co.uk": ['path',],
    "http://www.google.co.uk/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "https://www.google.co.uk/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "http://google.co.uk/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "https://google.co.uk/~as_db3.2123/134-1a": ['netloc','path','scheme',],
    "www.google.co.uk/~as_db3.2123/134-1a": ['path',],
    "google.co.uk/~as_db3.2123/134-1a": ['path',],
    "https://...": ['netloc', 'scheme',],
    "https://..": ['netloc', 'scheme',],
    "https://.": ['netloc', 'scheme',],
    "file:///mike/was/here": ['path','scheme',],
    "https://.google.com": ['netloc','scheme',],
    "https://..google.com": ['netloc','scheme',],
    "https://...google.com": ['netloc','scheme',],
    "https://.google..com": ['netloc','scheme',],
    "https://.google...com": ['netloc','scheme',],
    "https://...google..com": ['netloc','scheme',],
    "https://...google...com": ['netloc','scheme',],
    ".google.com": ['path',],
    ".google.co.": ['path',],
    "https://google.co.": ['netloc','scheme',],
}


def test_EnsureURL():
    with pytest.raises(ValueError):
        # only str input
        EnsureURL()(5)
    assert EnsureURL().short_description() == 'URL'
    assert EnsureURL(
        required=['scheme', 'netloc']
    ).short_description() == "URL with required ['scheme', 'netloc'] component(s)"
    assert EnsureURL(
        forbidden=['fragment']
    ).short_description() == "URL with no ['fragment'] component(s)"
    assert EnsureURL(
        # yes, it need not make sense
        required=['a'], forbidden=['b']
    ).short_description() == "URL with required ['a'] and with no ['b'] component(s)"

    any_url = EnsureURL()
    for tc in url_testcases.keys():
        any_url(tc)

    for t in ['netloc', 'path', 'scheme']:
        cnotag = EnsureURL(forbidden=[t])
        cnotag_parsed = EnsureParsedURL(forbidden=[t])
        for url, tags in url_testcases.items():
            if t in tags:
                with pytest.raises(ConstraintError) as e:
                    cnotag(url)
                assert f"forbidden '{t}'" in str(e.value)
            else:
                cnotag(url)
                cnotag_parsed(url)
        ctag = EnsureURL(required=[t])
        ctag_parsed = EnsureParsedURL(required=[t])
        for url, tags in url_testcases.items():
            if t not in tags:
                with pytest.raises(ConstraintError) as e:
                    ctag(url)
                assert f"missing '{t}'" in str(e.value)
            else:
                ctag(url)
                ctag_parsed(url)


def test_EnsureURL_match():
    # must contain a UUID
    c = EnsureURL(
        match='^.*([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}).*$',
    )
    with pytest.raises(ValueError):
        c('http://example.com')
    # it does not matter where it is
    for url in (
            'https://s.kg.eb.eu/i/a8932c7e-063c-4131-ab96-996d843998e9',
            'ssh://4ac9f0bc-560d-47e0-8916-7b24da9bb0ce.com/home',
    ):
        c(url)


def test_EnsureDataset(tmp_path, no_result_rendering):
    with pytest.raises(ValueError):
        # will not return a Dataset from sensless input
        EnsureDataset()(5)
    # by default the installation state is not checked
    # this matches the behavior of the original implementation
    # from datalad-core
    assert EnsureDataset()(tmp_path).ds.pathobj == tmp_path

    # this is same with auto-discovery of a dataset from CWD
    # (ie. with None as the argument)
    with chpwd(tmp_path):
        assert EnsureDataset()(None).ds.pathobj == tmp_path
        assert EnsureDataset(installed=False)(None).ds.pathobj == tmp_path
        # unless installation-verification is turned on
        with pytest.raises(NoDatasetFound):
            EnsureDataset(installed=True)(None)

    # any return value created from not-a-dataset-instance
    # has the original argument as an attribute
    assert EnsureDataset()(tmp_path).original == tmp_path

    # but it can be turned on, and then yields the specific
    # exception that datalad-core's require_dataset() would
    # give
    with pytest.raises(NoDatasetFound):
        EnsureDataset(installed=True)('/nothere_datalad_test')

    # we can also ensure absence
    assert EnsureDataset(installed=False)(tmp_path).ds.pathobj == tmp_path

    # absence detection with a dataset instance
    with pytest.raises(ValueError):
        EnsureDataset(installed=True)(
            # this provides the instance for testing
            EnsureDataset()(tmp_path).ds
        )

    #
    # tmp_path has a dataset from here
    #

    # create a dataset, making sure it did not exist before
    ds = EnsureDataset(installed=False)(tmp_path).ds.create()
    assert EnsureDataset()(ds).ds == ds
    assert EnsureDataset()(ds).original == ds

    # existence verified
    assert EnsureDataset(installed=True)(ds).ds.pathobj == tmp_path

    # check presence detection with path
    with pytest.raises(ValueError):
        EnsureDataset(installed=False)(tmp_path)
    # check presence detection and with dataset instance
    with pytest.raises(ValueError):
        EnsureDataset(installed=False)(ds)

    assert EnsureDataset().short_description() == '(path to) a Dataset'
    assert EnsureDataset(
        installed=True).short_description() == '(path to) an existing Dataset'
    assert EnsureDataset(
        installed=False).short_description() == \
        '(path to) a non-existing Dataset'

    # smoke test for idcheck:
    assert EnsureDataset(require_id=True)(ds).ds == ds
    assert EnsureDataset(require_id=False)(ds).ds == ds
    # unset the dataset ID to test whether an ID check would raise, but
    # bring it back later in case future tests need it
    id = ds.config.get('datalad.dataset.id')
    ds.config.unset('datalad.dataset.id', scope='branch')
    with pytest.raises(ValueError):
        EnsureDataset(require_id=True)(tmp_path)



from __future__ import annotations

# import all the pieces one would need for an implementation
# in a single place
from annexremote import UnsupportedRequest
from typing import Any

from datalad.customremotes import (
    # this is an enhanced RemoteError that self-documents its cause
    RemoteError,
    SpecialRemote as _SpecialRemote,
)
from datalad.customremotes.main import main as super_main

from datalad_next.datasets import LeanAnnexRepo


class SpecialRemote(_SpecialRemote):
    """Base class of all datalad-next git-annex special remotes"""
    def __init__(self, annex):
        super(SpecialRemote, self).__init__(annex=annex)

        self._repo = None
        self._remotename = None

    @property
    def repo(self) -> LeanAnnexRepo:
        """Returns a representation of the underlying git-annex repository

        An instance of :class:`~datalad_next.datasets.LeanAnnexRepo` is
        returned, which intentionally provides a restricted API only. In order
        to limit further proliferation of the ``AnnexRepo`` API.
        """
        if self._repo is None:
            self._repo = LeanAnnexRepo(self.annex.getgitdir())
        return self._repo

    @property
    def remotename(self) -> str:
        """Name of the (git) remote the special remote is operating under"""

        if self._remotename is None:
            self._remotename = self.annex.getgitremotename()
        return self._remotename

    def get_remote_gitcfg(
            self,
            remotetypename: str,
            name: str,
            default: Any | None = None,
            **kwargs
    ):
        """Get a particular Git configuration item for the special remote

        This target configuration here is *not* the git-annex native
        special remote configuration that is provided or altered with
        ``initremote`` and ``enableremote``, and is committed to the
        ``git-annex`` branch. Instead this is a clone and remote
        specific configuration, declared in Git's configuration system.

        The configuration items queried have the naming scheme::

            remote.<remotename>.<remotetypename>-<name>
            datalad.<remotetypename>.<name>

        where ``<remotename>`` is the name of the Git remote, the special
        remote is operating under, ``<remotetypename>`` is the name of the
        special remote implementation (e.g., ``uncurl``), and ``<name>``
        is the name of a particular configuration flavor.

        Parameters
        ----------
        remotetypename: str
          Name of the special remote implementation configuration is
          requested for.
        name: str
          The name of the "naked" configuration item, without any
          sub/sections. Must be a valid git-config variable name, i.e.,
          case-insensitive, only alphanumeric characters and -, and
          must start with an alphabetic character.
        default:
          A default value to be returned if there is no configuration.
        **kwargs:
          Passed on to :func:`datalad_next.config.ConfigManager.get()`

        Returns
        -------
        Any
          If a remote-specific configuration exists, it is reported. Otherwise
          a remote-type specific configuration is reported, or the default
          provided with the method call, if no configuration is found at all.
        """
        cfgget = self.repo.config.get
        return cfgget(
            f'remote.{self.remotename}.{remotetypename}-{name}',
            default=cfgget(
                f'datalad.{remotetypename}.{name}',
                default=default,
            )
        )



from pathlib import PurePosixPath
import pytest
import subprocess

from datalad.distributed.ora_remote import (
    RIARemoteError,
    SSHRemoteIO,
)


def test_sshremoteio(sshserver, tmp_path):
    sshurl, sshlocalpath = sshserver
    io = SSHRemoteIO(sshurl)
    # relative path, must be interpreted relative to given base url
    testfpath = 'dummy.txt'
    # we run in a tmp dir, test file must not exit
    assert not io.exists(testfpath)

    # TODO this content has a trailing newline, because "write_file()" requires
    # that. Madness. Remove when fixed, must work without.
    testcontent = 'two\nlines'
    io.write_file(testfpath, testcontent)
    # now we have a file
    assert io.exists(testfpath)
    # read content matches
    assert io.read_file(testfpath) == testcontent

    # create directory, make it interesting and have a space in the name
    testdirpath = 'something spacy'
    assert not io.exists(testdirpath)
    io.mkdir(testdirpath)
    assert io.exists(testdirpath)

    # download the testfile to local storage
    local_testfpath = tmp_path / testfpath
    # no progress callback
    io.get(testfpath, local_testfpath, lambda x: x)
    assert local_testfpath.read_text() == testcontent

    # upload to subdir
    testfpath_subdir = f'{testdirpath}/{testfpath}'
    assert not io.exists(testfpath_subdir)
    # TODO make absolutification unnecessary
    from urllib.parse import urlparse
    io.put(
        local_testfpath,
        f'{urlparse(sshurl).path}/{testfpath_subdir}',
        # no progress callback
        lambda x: x)
    assert io.exists(testfpath_subdir)

    # symlinks
    testfpath_link = 'dummy_link.txt'
    assert not io.exists(testfpath_link)
    io.symlink(testfpath, testfpath_link)
    assert io.exists(testfpath_link)
    assert io.read_file(testfpath_link) == testcontent

    # rename and delete
    # requires a Pure(Posix)Path object here
    io.rename(testfpath_subdir, PurePosixPath('deleteme'))
    assert not io.exists(testfpath_subdir)
    io.remove(PurePosixPath('deleteme'))
    assert not io.exists('deleteme')
    io.remove_dir(PurePosixPath(testdirpath))
    assert not io.exists(testdirpath)


def test_sshremoteio_7z(sshserver, tmp_path):
    sshurl, sshlocalpath = sshserver
    io = SSHRemoteIO(sshurl)
    # ensure we have a remote 7z
    if not io.get_7z():
        raise pytest.skip("No 7z available on SSH server target")

    testarchivefpath = 'my.7z'
    testfpath = 'dummy space.txt'
    testcontent = 'two\nlines\n'
    io.write_file(testfpath, testcontent)

    io.servershell(
        f'7z a "{testarchivefpath}" "{testfpath}"',
        check=True,
    )
    # we have an archive
    assert io.exists(testarchivefpath)
    # we have the test file in it
    assert io.in_archive(testarchivefpath, testfpath)
    # the "in" test means something
    assert not io.in_archive(testarchivefpath, "random_name")

    # we can pull from the archive
    extractfpath = tmp_path / 'extracted.txt'
    io.get_from_archive(testarchivefpath, testfpath, extractfpath, lambda x: x)
    assert extractfpath.read_text() == testcontent

    with pytest.raises(RIARemoteError):
        io.get_from_archive(
            'invalid_archive', testfpath, extractfpath, lambda x: x)
    with pytest.raises(RIARemoteError):
        io.get_from_archive(
            testarchivefpath, 'invalid_member', extractfpath, lambda x: x)



import warnings
from ..deprecate import deprecated

import pytest


@deprecated(msg='nothing to see here', version='1.0')
def deprecated_function(inputstring):
    return inputstring


class RandomClass(object):

    @deprecated(msg="nothing to see here", version='1.0')
    def deprecated_method(self, inputstring):
        return inputstring


@deprecated(msg="nothing to see here", version='1.0')
class DeprecatedClass(object):

    def __call__(self, inputstring):
        return inputstring


@deprecated(msg='nothing to see here', kwarg='inputmode', version='1.0')
def deprecated_function_param(inputmode='default', other_param=None):
    return inputmode


class RandomClassParam(object):

    @deprecated(msg="nothing to see here", kwarg='inputmode', version='1.0')
    def deprecated_method(self, inputmode='default', other_param=None):
        return inputmode


@deprecated(msg='nothing to see here', kwarg='inputmode',
            kwarg_values=['default'], version='1.0')
def deprecated_function_param_value(inputmode='default'):
    return inputmode


@deprecated(msg='nothing to see here', kwarg='inputmode',
            kwarg_values=[None], version='1.0')
def deprecated_function_param_value_none(inputmode=None):
    return inputmode


class RandomClassParamValue(object):

    @deprecated(msg="nothing to see here", kwarg='inputmode',
                kwarg_values=['default'], version='1.0')
    def deprecated_method(self, inputmode='default'):
        return inputmode


@deprecated(msg='nothing to see here', version='1.0', kwarg='mode')
@deprecated(msg='even less to see here', version='1.0', kwarg='othermode')
def double_deprecated_function(mode='default', othermode='moredefault'):
    return (mode, othermode)


@deprecated(msg='nothing to see here', version='1.0', kwarg='mode',
            kwarg_values=['1', '2'])
def two_deprecated_values(mode='default'):
    return mode


def test_deprecated():
    # deprecations for entire functions/classes
    input_string = 'hello world'
    for func in [deprecated_function,
                 RandomClass().deprecated_method]:
        with pytest.warns(
                DeprecationWarning,
                match=f"{func.__module__}.{func.__name__} was deprecated in version 1.0. nothing to see here"):
            res = func(inputstring=input_string)
            assert res == input_string

    with pytest.warns(DeprecationWarning, match="nothing to see here"):
        DeprecatedClass()

    # deprecations for a kwarg
    inputmode = 'default'
    for func in [deprecated_function_param,
                 RandomClassParam().deprecated_method]:
        with pytest.warns(DeprecationWarning, match="argument 'inputmode'"):
            res = func(inputmode=inputmode)
            assert res == inputmode

    # deprecations for a kwarg value
    for func in [deprecated_function_param_value,
                 RandomClassParamValue().deprecated_method,
                 ]:
        with pytest.warns(
                DeprecationWarning,
                match="Use of values {'default'} for argument 'inputmode'"):
            res = func(inputmode=inputmode)
            assert res == inputmode

    # `None` value deprecation is supported
    # test in many complicated forms
    for v in (None, (None,), [None], {None: 'some'}):
        with pytest.warns(
                DeprecationWarning,
                match="Use of values {None} for argument 'inputmode'"):
            res = deprecated_function_param_value_none(inputmode=v)
            assert res == v

    # no deprecations for an unused deprecated parameter or parameter value
    for func in [deprecated_function_param_value,
                 RandomClassParamValue().deprecated_method,
                 ]:
        with warnings.catch_warnings(record=True) as record:
            res = func(inputmode='not-deprecated')
        assert res == 'not-deprecated'
        assert len(record) == 0

        for func in [deprecated_function_param,
                     RandomClassParam().deprecated_method]:
            with warnings.catch_warnings(record=True) as record:
                res = func(other_param='something!')
            assert res == inputmode
            assert len(record) == 0

    # make sure it catches the parameter even if its a list
    for func in [deprecated_function_param_value,
                 RandomClassParamValue().deprecated_method,
                 ]:
        with pytest.warns(
                DeprecationWarning,
                match="Use of values {'default'} for argument 'inputmode'"):
            res = func(inputmode=[inputmode])
            assert res == [inputmode]
        with warnings.catch_warnings(record=True) as record:
            res = func(inputmode=['not-deprecated'])
        assert res == ['not-deprecated']
        assert len(record) == 0

    # two decorators work as expected
    with pytest.warns(DeprecationWarning) as record:
        res = double_deprecated_function(mode='1', othermode='2')
        assert res == ('1', '2')
        assert len(record) == 2
        assert 'nothing to see here' in str(record.list[0].message)
        assert 'even less to see here' in str(record.list[1].message)

    # test that everything works when the function has several deprecated values

    with pytest.warns(DeprecationWarning):
        res = two_deprecated_values(mode='1')
        assert res == '1'
        # shouldn't matter if the parameter value is a list
        res = two_deprecated_values(mode=['1'])
        assert res == ['1']
    with warnings.catch_warnings(record=True) as record:
        res = two_deprecated_values(mode='safe')
    assert res == 'safe'
    assert len(record) == 0



import os
from pathlib import PurePath
import pytest

from datalad_next.tests import (
    create_tree,
)
from datalad_next.utils import (
    check_symlink_capability,
    rmtree,
)

from ..directory import (
    DirectoryItem,
    FileSystemItemType,
    iter_dir,
)
from ..utils import compute_multihash_from_fp


@pytest.fixture(scope="function")
def dir_tree(tmp_path_factory):
    path = tmp_path_factory.mktemp("dir_tree")
    create_tree(
        path,
        {
            "random_file1.txt": "some content",
            "some_dir": {
                "file_in_dir.txt": "some content in file in dir",
            },
        }
    )
    symlink = path / 'symlink'
    symlink_target = path / 'some_dir' / "file_in_dir.txt"

    if check_symlink_capability(symlink, symlink_target):
        symlink.symlink_to(symlink_target)

    yield path
    rmtree(path)


def test_iter_dir(dir_tree):
    target_hash = dict(md5='9893532233caff98cd083a116b013c0b',
                       SHA1='94e66df8cd09d410c62d9e0dc59d3a884e458e05')
    target_paths = [
        (dir_tree / 'random_file1.txt', FileSystemItemType.file, {}),
        (dir_tree / 'some_dir', FileSystemItemType.directory, {}),
    ]
    if check_symlink_capability(dir_tree / '__dummy1__',
                                dir_tree / '__dummy2__'):
        target_paths.append((
            dir_tree / 'symlink', FileSystemItemType.symlink,
            # how `readlink()` behaves on windows is fairly complex
            # rather than anticipating a result (that changes with
            # python version, see https://bugs.python.org/issue42957),
            # we simply test that this is compatible with `os.readlink()`
            dict(link_target=PurePath(os.readlink(dir_tree / 'symlink'))),
        ))
    target = [
        DirectoryItem(
            name=PurePath(path.name),
            type=type,
            size=path.lstat().st_size,
            mode=path.lstat().st_mode,
            mtime=path.lstat().st_mtime,
            uid=path.lstat().st_uid,
            gid=path.lstat().st_gid,
            **kwa
        )
        for path, type, kwa in target_paths
    ]

    iter_dir_res = []
    for i in iter_dir(dir_tree, fp=True):
        if i.fp:
            # capitalization of algorithm labels is preserved
            assert compute_multihash_from_fp(
                i.fp, ['md5', 'SHA1']) == target_hash
            # we null the file pointers to ease the comparison
            i.fp = None
        iter_dir_res.append(i)
    assert len(iter_dir_res) == len(target)

    # check iter_dir() to be robust to concurrent removal
    it = iter_dir(dir_tree)
    # start iteration
    next(it)
    # wipe out content
    for i in dir_tree.glob('*'):
        rmtree(i)
    # consume the rest of the generator, nothing more, but also no crashing
    assert [] == list(it)



"""UI abstractions for user communication

.. currentmodule:: datalad_next.uis
.. autosummary::
   :toctree: generated

   ansi_colors
   ui_switcher
"""

# make more obvious that this is a frontend that behaves
# differently depending on many conditions
from datalad.ui import ui as ui_switcher

from datalad.support import ansi_colors



"""Report on the content of a Git-annex repository worktree

The main functionality is provided by the :func:`iter_annexworktree()`
function.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from more_itertools import intersperse
from pathlib import (
    Path,
    PurePath,
)
from typing import (
    Dict,
    Type,
    Union,
    Any,
    Generator,
)
from datasalad.itertools import (
    itemize,
    load_json,
    route_in,
    route_out,
    StoreOnly,
)

from datalad_next.consts import on_windows
from datalad_next.repo_utils import has_initialized_annex
from datalad_next.runners import iter_git_subproc

from .gitworktree import (
    GitWorktreeItem,
    GitWorktreeFileSystemItem,
    iter_gitworktree,
)
from .utils import FileSystemItemType


lgr = logging.getLogger('datalad.ext.next.iter_collections.annexworktree')


@dataclass
class AnnexWorktreeItem(GitWorktreeItem):
    annexkey: str | None = None
    annexsize: int | None = None
    # annex object path, relative to the item
    annexobjpath: PurePath | None = None

    @classmethod
    def from_gitworktreeitem(
        cls,
        item: GitWorktreeItem,
    ) -> "AnnexWorktreeItem":
        return cls(**item.__dict__)


@dataclass
class AnnexWorktreeFileSystemItem(GitWorktreeFileSystemItem):
    annexkey: str | None = None
    annexsize: int | None = None
    # annex object path, relative to the item
    annexobjpath: PurePath | None = None


# TODO this iterator should get a filter mechanism to limit it to a single
# directory (non-recursive). This will be needed for gooey.
# unlike iter_gitworktree() we pay a larger dedicated per item cost.
# Given that the switch to iterative processing is also made for
# iter_gitworktree() we should provide the same filtering for that one
# too!
def iter_annexworktree(
    path: Path,
    *,
    untracked: str | None = 'all',
    link_target: bool = False,
    fp: bool = False,
    recursive: str = 'repository',
) -> Generator[AnnexWorktreeItem | AnnexWorktreeFileSystemItem, None, None]:
    """Companion to ``iter_gitworktree()`` for git-annex repositories

    This iterator wraps
    :func:`~datalad_next.iter_collections.gitworktree.iter_gitworktree`.
    For each item, it determines whether it is an annexed file. If so,
    it amends the yielded item with information on the respective
    annex key, the byte size of the key, and its (would-be) location
    in the repository's annex.

    The basic semantics of all arguments are identical to
    :func:`~datalad_next.iter_collections.gitworktree.iter_gitworktree`.
    Importantly, with ``fp=True``, an annex object is opened directly,
    if available. If not available, no attempt is made to open the associated
    symlink or pointer file.

    With ``link_target`` and ``fp`` disabled items of
    type :class:`AnnexWorktreeItem` are yielded, otherwise
    :class:`AnnexWorktreeFileSystemItem` instances are yielded. In both cases,
    ``annexkey``, ``annexsize``, and ``annnexobjpath`` properties are provided.

    .. note::
      Although ``annexobjpath`` is always set for annexed content, that does
      not imply that an object at this path actually exists. The latter will
      only be the case if the annexed content is present in the work tree,
      typically as a result of a `datalad get`- or `git annex get`-call.

    Parameters
    ----------
    path: Path
      Path of a directory in a git-annex repository to report on. This
      directory need not be the root directory of the repository, but
      must be part of the repository's work tree.
    untracked: {'all', 'whole-dir', 'no-empty-dir'} or None, optional
      If not ``None``, also reports on untracked work tree content.
      ``all`` reports on any untracked file; ``whole-dir`` yields a single
      report for a directory that is entirely untracked, and not individual
      untracked files in it; ``no-empty-dir`` skips any reports on
      untracked empty directories.
    link_target: bool, optional
      If ``True``, information matching a
      :class:`~datalad_next.iter_collections.utils.FileSystemItem`
      will be included for each yielded item, and the targets of
      any symlinks will be reported, too.
    fp: bool, optional
      If ``True``, information matching a
      :class:`~datalad_next.iter_collections.utils.FileSystemItem`
      will be included for each yielded item, but without a
      link target detection, unless ``link_target`` is given.
      Moreover, each file-type item includes a file-like object
      to access the file's content. This file handle will be closed
      automatically when the next item is yielded.
    recursive: {'repository', 'no'}, optional
      Pass on to
      :func:`~datalad_next.iter_collections.gitworktree.iter_gitworktree`,
      thereby determining which items this iterator will yield.

    Yields
    ------
    :class:`AnnexWorktreeItem` or :class:`AnnexWorktreeFileSystemItem`
      The ``name`` attribute of an item is a ``PurePath`` instance with
      the corresponding (relative) path, in platform conventions.
    """

    glsf = iter_gitworktree(
        path,
        untracked=untracked,
        link_target=False,
        fp=False,
        recursive=recursive,
    )

    if not has_initialized_annex(path):
        # this is not an annex repo.
        # we just yield the items from the gitworktree iterator.
        # we funnel them through the standard result item prep
        # function for type equality.
        # when a recursive-mode other than 'repository' will be
        # implemented, this implementation needs to be double-checked
        # to avoid decision making on submodules just based on
        # the nature of the toplevel repo.
        for item in glsf:
            yield _get_worktree_item(
                path, get_fs_info=link_target, git_item=item)
        return

    git_fileinfo_store: list[Any] = list()
    # this is a technical helper that will just store a bunch of `None`s
    # for aligning item-results between git-ls-files and git-annex-find
    _annex_git_align: list[Any] = list()

    with \
            iter_git_subproc(
                # we get the annex key for any filename
                # (or empty if not annexed)
                ['annex', 'find', '--anything', '--format=${key}\\n',
                 '--batch'],
                # intersperse items with newlines to trigger a batch run
                # this avoids string operations to append newlines to items
                input=intersperse(
                    b'\n',
                    # use `GitWorktree*`-elements yielded by `iter_gitworktree`
                    # to create an `AnnexWorktreeItem` or
                    # `AnnexWorktreeFileSystemItem` object, which is stored in
                    # `git_fileinfo_store`. Yield a string representation of
                    # the path contained in the `GitWorktree*`-element yielded
                    # by `iter_gitworktree`
                    route_out(
                        glsf,
                        git_fileinfo_store,
                        lambda git_worktree_item: (
                            str(git_worktree_item.name).encode(),
                            git_worktree_item
                        )
                    )
                ),
                cwd=path,
            ) as gaf, \
            iter_git_subproc(
                # get the key properties JSON-lines style
                ['annex', 'examinekey', '--json', '--batch'],
                # use only non-empty keys as input to `git annex examinekey`.
                input=intersperse(
                    # Add line ending to submit the key to batch processing in
                    # `git annex examinekey`.
                    b'\n',
                    route_out(
                        itemize(
                            gaf,
                            # although we declare a specific key output format
                            # for the git-annex find call, versions of
                            # git-annex <10.20231129 on Windows will terminate
                            # lines with '\r\n' instead of '\n'. We therefore use
                            # `None` as separator, which enables `itemize()`
                            # to use either separator, i.e. '\r\n' or '\n'.
                            sep=None if on_windows else b'\n',
                        ),
                        # we need this route-out solely for the purpose
                        # of maintaining a 1:1 relationship of items reported
                        # by git-ls-files and git-annex-find (merged again
                        # in the `route-in` that gives `results` below). The
                        # "store" here does not actually store anything other
                        # than`None`s (because the `key` --which is consumed by
                        # `git annex examinekey`-- is also present in the
                        # output of `git annex examinekey`).
                        _annex_git_align,
                        # do not process empty key lines. Non-empty key lines
                        # are processed, but nothing needs to be stored because
                        # the processing result includes the key itself.
                        lambda key: (key if key else StoreOnly, None)
                    )
                ),
                cwd=path,
            ) as gek:

        results = route_in(
            # the following `route_in` yields processed keys for annexed
            # files and `StoreOnly` for non-annexed files. Its
            # cardinality is the same as the cardinality of
            # `iter_gitworktree`, i.e. it produces data for each element
            # yielded by `iter_gitworktree`.
            route_in(
                load_json(itemize(gek, sep=None)),
                _annex_git_align,
                # `processed` data is either `StoreOnly` or detailed
                # annex key information. we just return `process_data` as
                # result, because `join_annex_info` knows how to incorporate
                # it into an `AnnexWorktree*`-object.
                lambda processed_data, _: processed_data
            ),
            git_fileinfo_store,
            _join_annex_info,
        )

        # at this point, each item in `results` is a dict with a `git_item`
        # key that hold a `GitWorktreeItem` instance, plus additional annex
        # related keys added by join_annex_info() for annexed files
        if not fp:
            # life is simpler here, we do not need to open any files in the
            # annex, hence all processing can be based in the information
            # collected so far
            for res in results:
                yield _get_worktree_item(path, get_fs_info=link_target, **res)
            return

        # if we get here, this is about file pointers...
        # for any annexed file we need to open, we need to locate it in
        # the annex. we get `annexobjpath` in the results. this is
        # relative to `path`. We could not use the `link_target`, because
        # we might be in a managed branch without link.
        path = Path(path)
        for res in results:
            try:
                item = _get_worktree_item(path, get_fs_info=True, **res)
            except FileNotFoundError:
                # there is nothing to open, yield non FS item
                item = _get_worktree_item(path, get_fs_info=False, **res)
                yield item
                continue

            # determine would file we would open
            fp_src = None
            if item.annexobjpath is not None:
                # this is an annexed file
                fp_src = item.annexobjpath
            elif item.type == FileSystemItemType.file \
                    and item.annexkey is None:
                # regular file (untracked or tracked)
                fp_src = item.name
            elif item.type == FileSystemItemType.symlink \
                    and item.annexkey is None:
                # regular symlink
                fp_src = item.name
            if fp_src is None:
                # nothing to open
                yield item
            else:
                fp_src_fullpath = path / fp_src
                if not fp_src_fullpath.exists():
                    # nothing there to open (would resolve through a symlink)
                    yield item
                else:
                    with fp_src_fullpath.open('rb') as active_fp:
                        item.fp = active_fp
                        yield item


def _get_worktree_item(
    base_path: Path,
    get_fs_info: bool,
    git_item: GitWorktreeItem,
    annexkey: str | None = None,
    annexsize: int | None = None,
    annexobjpath: str | None = None,
) -> AnnexWorktreeFileSystemItem | AnnexWorktreeItem:
    """Internal helper to get an item from ``_join_annex_info()`` output

    The assumption is that minimal investigations have been done
    until this helper is called. In particular, no file system inspects
    have been performed.

    Depending on whether a user requested file system information to be
    contained in the items (``get_fs_info``), either
    ``AnnexWorktreeFileSystemItem`` or ``AnnexWorktreeItem`` is returned.

    The main workhorse of this function if
    ``AnnexWorktreeFileSystemItem.from_path()``. Besides calling it,
    information is only taken from arguments and injected into the item
    instances.
    """
    # we did not do any filesystem inspection previously, so
    # do now when link_target is enabled
    item = AnnexWorktreeFileSystemItem.from_path(
        base_path / git_item.name,
        link_target=True,
    ) if get_fs_info else AnnexWorktreeItem.from_gitworktreeitem(git_item)
    # amend the AnnexWorktree* object with the available git info
    item.gitsha = git_item.gitsha
    item.gittype = git_item.gittype
    # amend the AnnexWorktree* object with the available annex info
    item.annexkey = annexkey
    item.annexsize = annexsize
    item.annexobjpath = annexobjpath
    return item


def _join_annex_info(
    processed_data: Union[Type[StoreOnly], Dict[str, str]],
    stored_data: GitWorktreeItem,
) -> dict:
    """Internal helper to join results from pipeline stages

    All that is happening here is that information from git and git-annex
    inquiries gets merged into a single result dict.
    """
    joined = dict(git_item=stored_data)
    if processed_data is StoreOnly:
        # this is a non-annexed item, nothing to join
        return joined
    else:
        # here processed data is a dict with properties from annex examinekey
        joined.update(
            annexkey=processed_data['key'],
            annexsize=int(processed_data['bytesize']),
            annexobjpath=PurePath(str(processed_data['objectpath'])),
        )
        return joined
